# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains all the different availability states that could apply to all of the language server features.

The actual availability of the each individual feature is handled by the vscode client
and is passed to the pyre client entry point (pyre.py) upon initialization of the persistent client.
"""

from __future__ import annotations

import dataclasses
import enum


class _Availability(enum.Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"

    @staticmethod
    def from_enabled(enabled: bool) -> _Availability:
        return _Availability.ENABLED if enabled else _Availability.DISABLED

    def is_enabled(self) -> bool:
        return self == _Availability.ENABLED

    def is_disabled(self) -> bool:
        return self == _Availability.DISABLED


class TypeCoverageAvailability(enum.Enum):
    DISABLED = "disabled"
    FUNCTION_LEVEL = "function_level"
    EXPRESSION_LEVEL = "expression_level"


# User-facing features
StatusUpdatesAvailability = _Availability
TypeErrorsAvailability = _Availability
UnsavedChangesAvailability = _Availability

# Telemetry: is the editor able to forward events somewhere?
TelemetryAvailability = _Availability


@dataclasses.dataclass(frozen=True)
class LanguageServerFeatures:
    status_updates: StatusUpdatesAvailability = StatusUpdatesAvailability.ENABLED
    type_coverage: TypeCoverageAvailability = TypeCoverageAvailability.DISABLED
    type_errors: TypeErrorsAvailability = TypeErrorsAvailability.ENABLED
    unsaved_changes: UnsavedChangesAvailability = UnsavedChangesAvailability.DISABLED
    telemetry: TelemetryAvailability = TelemetryAvailability.DISABLED



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
from pathlib import Path
from typing import List

from pyre_extensions import override

from ..configuration import Configuration
from ..filesystem import path_exists
from ..repository import Repository
from .command import Command, CommandArguments, ErrorSource
from .fixme import Fixme

LOG: logging.Logger = logging.getLogger(__name__)


class GlobalVersionUpdate(Command):
    def __init__(
        self,
        *,
        repository: Repository,
        error_source: str,
        hash: str,
        paths: List[Path],
        no_commit: bool,
    ) -> None:
        super().__init__(repository)
        self._error_source: str = error_source
        self._hash: str = hash
        self._paths: List[Path] = paths
        self._no_commit: bool = no_commit

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "GlobalVersionUpdate":
        return GlobalVersionUpdate(
            repository=repository,
            error_source=arguments.error_source,
            hash=arguments.hash,
            paths=arguments.paths,
            no_commit=arguments.no_commit,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(GlobalVersionUpdate, GlobalVersionUpdate).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument("hash", help="Hash of new Pyre version")
        parser.add_argument(
            "--paths",
            nargs="*",
            help="A list of paths to local Pyre projects.",
            default=[],
            type=path_exists,
        )
        parser.add_argument(
            "--error-source",
            type=ErrorSource,
            choices=list(ErrorSource),
            default=ErrorSource.GENERATE,
        )
        parser.add_argument(
            "--no-commit", action="store_true", help="Keep changes in working state."
        )

    def _set_local_overrides(
        self, configuration_paths: List[Path], old_version: str
    ) -> None:
        for configuration_path in configuration_paths:
            if "mock_repository" in str(configuration_path):
                # Skip local configurations we have for testing.
                continue
            local_configuration = Configuration(configuration_path)
            if local_configuration.version:
                LOG.info(
                    "Skipping %s as it already has a custom version field.",
                    configuration_path,
                )
                continue
            local_configuration.set_version(old_version)
            local_configuration.write()

    def _suppress_global_errors(self, global_configuration: Configuration) -> None:
        if global_configuration.targets or global_configuration.source_directories:
            LOG.info("Suppressing errors after upgrading global version.")
            command_arguments = CommandArguments(
                comment=None,
                max_line_length=None,
                truncate=False,
                unsafe=False,
                force_format_unsuppressed=False,
                lint=True,
                no_commit=True,
                should_clean=True,
            )
            fixme_command = Fixme(
                command_arguments,
                repository=self._repository,
                error_source=self._error_source,
            )
            fixme_command.run()

        self._repository.commit_changes(
            commit=(not self._no_commit),
            title="Update pyre global configuration version",
            summary=f"Automatic upgrade to hash `{self._hash}`",
            ignore_failures=True,
        )

    @override
    def run(self) -> None:
        global_configuration = Configuration.find_project_configuration()

        # Update to new global version.
        configuration = Configuration(global_configuration)
        old_version = configuration.version
        if not old_version:
            LOG.error(
                "Global configuration at %s has no version field.", global_configuration
            )
            return
        configuration.set_version(self._hash)
        configuration.write()

        paths = self._paths
        configuration_paths = (
            [path / ".pyre_configuration.local" for path in paths]
            if paths
            else [
                configuration.get_path()
                for configuration in Configuration.gather_local_configurations()
            ]
        )

        self._set_local_overrides(configuration_paths, old_version)
        self._suppress_global_errors(configuration)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module represents the top level class and entry point for the LSP client
to interact with the Pyre Server. The LSP client will invoke the top level
method here, which will instantiate a running instance of a persistent Pyre
client that will interact with the pyre server.

The meaning of "persistent" in this context is to illustrate that the
connection to the pyre server is non-transient and is designed to stay alive
during the entire duration of a user's interaction with Vscode.

The main responsibilities of this class is to:

-   Instantiate an instance of the persistent Pyre client and ensure
    re-launch of the Pyre server in the scenario that it goes down.

"""

from __future__ import annotations

import asyncio
import logging
import os
import traceback
from typing import Optional

from .. import backend_arguments, background_tasks, log_lsp_event, timer, version
from ..language_server import connections, features, protocol as lsp
from . import (
    daemon_querier,
    incremental,
    launch_and_subscribe_handler,
    pyre_language_server,
    pyre_server_options,
    server_state as state,
    status_message_handler,
    subscription,
    type_error_handler,
)
from .initialization import async_try_initialize_loop, InitializationExit
from .server_state import ServerState

LOG: logging.Logger = logging.getLogger(__name__)

COMMAND_NAME = "persistent"


def process_initialize_request(
    parameters: lsp.InitializeParameters,
    language_server_features: Optional[features.LanguageServerFeatures] = None,
) -> lsp.InitializeResult:
    LOG.info(
        f"Received initialization request from {parameters.client_info} "
        f" (pid = {parameters.process_id})"
    )
    if language_server_features is None:
        language_server_features = features.LanguageServerFeatures()
    server_info = lsp.Info(name="pyre", version=version.__version__)
    server_capabilities = lsp.ServerCapabilities(
        text_document_sync=lsp.TextDocumentSyncOptions(
            open_close=True,
            change=lsp.TextDocumentSyncKind.FULL,
            save=lsp.SaveOptions(include_text=False),
        ),
    )
    return lsp.InitializeResult(
        capabilities=server_capabilities, server_info=server_info
    )


READY_MESSAGE: str = "Pyre has completed an incremental check and is currently watching on further source changes."


class PyrePersistentSubscriptionResponseParser(
    launch_and_subscribe_handler.PyreSubscriptionResponseParser
):
    def parse_response(self, response: str) -> subscription.Response:
        return subscription.Response.parse(response)


class PyrePersistentDaemonLaunchAndSubscribeHandler(
    launch_and_subscribe_handler.PyreDaemonLaunchAndSubscribeHandler
):
    def __init__(
        self,
        server_options_reader: pyre_server_options.PyreServerOptionsReader,
        server_state: ServerState,
        client_status_message_handler: status_message_handler.ClientStatusMessageHandler,
        client_type_error_handler: type_error_handler.ClientTypeErrorHandler,
        remote_logging: Optional[backend_arguments.RemoteLogging] = None,
    ) -> None:
        super().__init__(
            server_options_reader,
            server_state,
            client_status_message_handler,
            client_type_error_handler,
            PyrePersistentSubscriptionResponseParser(),
            remote_logging,
        )

    async def handle_type_error_event(
        self,
        type_error_subscription: subscription.TypeErrors,
    ) -> None:
        await self.client_type_error_handler.clear_type_errors_for_client()
        self.client_type_error_handler.update_type_errors(
            type_error_subscription.errors
        )
        self.server_state.status_tracker.set_status(state.ConnectionStatus.READY)
        await self.client_type_error_handler.show_type_errors_to_client()
        await self.client_status_message_handler.log_and_show_status_message_to_client(
            READY_MESSAGE,
            short_message=f"{self.server_state.server_options.flavor.simple_name()} Ready",
            level=lsp.MessageType.INFO,
        )

    async def handle_status_update_event(
        self,
        status_update_subscription: subscription.StatusUpdate,
    ) -> None:
        flavor_simple_name = self.server_state.server_options.flavor.simple_name()
        if self.get_type_errors_availability().is_enabled():
            await self.client_type_error_handler.clear_type_errors_for_client()
        if status_update_subscription.kind == "Rebuilding":
            self.server_state.status_tracker.set_status(
                state.ConnectionStatus.BUCK_BUILDING
            )
            await self.client_status_message_handler.log_and_show_status_message_to_client(
                "Pyre is busy rebuilding the project for type checking...",
                short_message=f"{flavor_simple_name} (waiting for Buck)",
                level=lsp.MessageType.WARNING,
            )
        elif status_update_subscription.kind == "Rechecking":
            self.server_state.status_tracker.set_status(
                state.ConnectionStatus.INCREMENTAL_CHECK
            )
            await self.client_status_message_handler.log_and_show_status_message_to_client(
                "Pyre is busy re-type-checking the project...",
                short_message=f"{flavor_simple_name} (checking)",
                level=lsp.MessageType.WARNING,
            )
        elif status_update_subscription.kind == "Ready":
            self.server_state.status_tracker.set_status(state.ConnectionStatus.READY)
            await self.client_status_message_handler.log_and_show_status_message_to_client(
                READY_MESSAGE,
                short_message=f"{flavor_simple_name} Ready",
                level=lsp.MessageType.INFO,
            )

    async def _subscribe_to_type_errors(
        self,
        server_input_channel: connections.AsyncTextReader,
        server_output_channel: connections.AsyncTextWriter,
    ) -> None:
        subscription_name = f"persistent_{os.getpid()}"
        await server_output_channel.write(
            f'["SubscribeToTypeErrors", "{subscription_name}"]\n'
        )
        first_response = await launch_and_subscribe_handler.PyreDaemonLaunchAndSubscribeHandler._read_server_response(
            server_input_channel
        )
        initial_type_errors = incremental.parse_type_error_response(
            first_response
        ).errors
        self.client_type_error_handler.update_type_errors(initial_type_errors)
        await self.client_type_error_handler.show_type_errors_to_client()
        await self._run_subscription_loop(
            subscription_name,
            server_input_channel,
            server_output_channel,
        )

    async def _subscribe_to_state_changes(
        self,
        server_input_channel: connections.AsyncTextReader,
        server_output_channel: connections.AsyncTextWriter,
    ) -> None:
        subscription_name = f"persistent_{os.getpid()}"
        await server_output_channel.write(
            f'["SubscribeToStateChanges", "{subscription_name}"]\n'
        )
        await self._read_server_response(server_input_channel)
        await self._run_subscription_loop(
            subscription_name,
            server_input_channel,
            server_output_channel,
        )

    async def _subscribe(
        self,
        server_input_channel: connections.AsyncTextReader,
        server_output_channel: connections.AsyncTextWriter,
    ) -> None:
        if self.get_type_errors_availability().is_enabled():
            await self._subscribe_to_type_errors(
                server_input_channel,
                server_output_channel,
            )
        else:
            await self._subscribe_to_state_changes(
                server_input_channel,
                server_output_channel,
            )

    async def client_setup(self) -> None:
        """Persistent server does not require any client-side setup"""
        pass

    async def client_teardown(self) -> None:
        """Persistent server does not require any client-side teardown"""
        pass


async def run_persistent(
    server_options_reader: pyre_server_options.PyreServerOptionsReader,
    remote_logging: Optional[backend_arguments.RemoteLogging],
) -> int:
    initial_server_options = pyre_server_options.read_server_options(
        server_options_reader, remote_logging
    )
    stdin, stdout = await connections.create_async_stdin_stdout()

    initialize_result = await async_try_initialize_loop(
        stdin,
        stdout,
        remote_logging,
        compute_initialize_result=lambda parameters: process_initialize_request(
            parameters, initial_server_options.language_server_features
        ),
    )
    if isinstance(initialize_result, InitializationExit):
        return 0

    client_info = initialize_result.client_info
    log_lsp_event.log(
        remote_logging=remote_logging,
        event=log_lsp_event.LSPEvent.INITIALIZED,
        normals=(
            {}
            if client_info is None
            else {
                "lsp client name": client_info.name,
                "lsp client version": client_info.version,
            }
        ),
    )

    client_capabilities = initialize_result.client_capabilities
    LOG.debug(f"Client capabilities: {client_capabilities}")
    server_state = state.ServerState(
        client_capabilities=client_capabilities,
        server_options=initial_server_options,
    )
    querier = daemon_querier.PersistentDaemonQuerier(
        server_state=server_state,
    )
    client_type_error_handler = type_error_handler.ClientTypeErrorHandler(
        stdout, server_state, remote_logging
    )

    server = pyre_language_server.PyreLanguageServerDispatcher(
        input_channel=stdin,
        output_channel=stdout,
        server_state=server_state,
        daemon_manager=background_tasks.TaskManager(
            PyrePersistentDaemonLaunchAndSubscribeHandler(
                server_options_reader=server_options_reader,
                remote_logging=remote_logging,
                server_state=server_state,
                client_status_message_handler=status_message_handler.ClientStatusMessageHandler(
                    stdout, server_state
                ),
                client_type_error_handler=client_type_error_handler,
            )
        ),
        api=pyre_language_server.PyreLanguageServer(
            output_channel=stdout,
            server_state=server_state,
            querier=querier,
            client_type_error_handler=client_type_error_handler,
        ),
    )
    return await server.run()


def run(
    read_server_options: pyre_server_options.PyreServerOptionsReader,
    remote_logging: Optional[backend_arguments.RemoteLogging],
) -> int:
    command_timer = timer.Timer()
    error_message: Optional[str] = None
    try:
        return asyncio.get_event_loop().run_until_complete(
            run_persistent(
                read_server_options,
                remote_logging,
            )
        )
    except Exception:
        error_message = traceback.format_exc()
        LOG.exception("Uncaught error in persistent.run")
        return 1
    finally:
        log_lsp_event.log(
            remote_logging,
            log_lsp_event.LSPEvent.STOPPED,
            integers={"duration": int(command_timer.stop_in_millisecond())},
            normals={
                **({"exception": error_message} if error_message is not None else {})
            },
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
The `client.commands` namespace of Pyre contains the entrypoints
for all of the subcommands of pyre such as `pyre check` and
`pyre incremental`. It also contains some helper modules with shared
library code used by several commands.
"""

from . import (  # noqa F401
    analyze,
    check,
    incremental,
    infer,
    info,
    initialization,
    initialize,
    initialize_pysa,
    kill,
    persistent,
    profile,
    pyre_language_server,
    pysa_server,
    query,
    rage,
    report,
    report_any_expressions,
    restart,
    servers,
    start,
    statistics,
    stop,
    validate_models,
)
from .commands import ClientException as ClientException, ExitCode as ExitCode  # noqa



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""


class UserError(Exception):
    pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Core logic for the `pyre stop` command, which attempts to do a clean
shutdown of a running Pyre daemon by sending it a shutdown request
via a socket connection.
"""

import logging
from pathlib import Path

from .. import daemon_socket, frontend_configuration, identifiers
from ..language_server import connections
from . import commands

LOG: logging.Logger = logging.getLogger(__name__)


def stop_message(flavor: identifiers.PyreFlavor) -> str:
    if flavor not in (
        identifiers.PyreFlavor.CLASSIC,
        identifiers.PyreFlavor.SHADOW,
    ):
        raise AssertionError(
            f"Attempted to stop a server for unsupported flavor {flavor}"
        )
    return '["Stop"]'


def stop_server(socket_path: Path, flavor: identifiers.PyreFlavor) -> None:
    with connections.connect(socket_path) as (
        input_channel,
        output_channel,
    ):
        output_channel.write(f"{stop_message(flavor)}\n")
        # Wait for the server to shutdown on its side
        input_channel.read()


def remove_socket_if_exists(socket_path: Path) -> None:
    try:
        socket_path.unlink()
    except FileNotFoundError:
        pass
    except OSError as error:
        LOG.warning(f"Failed to remove socket file at `{socket_path}`: {error}")
    try:
        socket_path.with_suffix(socket_path.suffix + ".lock").unlink()
    except FileNotFoundError:
        pass
    except OSError as error:
        LOG.warning(f"Failed to remove lock file at `{socket_path}.lock`: {error}")


def run(
    configuration: frontend_configuration.Base, flavor: identifiers.PyreFlavor
) -> commands.ExitCode:
    socket_path = daemon_socket.get_socket_path(
        configuration.get_project_identifier(),
        flavor,
    )
    try:
        LOG.info(f"Stopping {flavor.value} server...")
        LOG.info(f"Socket is {socket_path}")
        stop_server(socket_path, flavor)
        LOG.info(
            f"Stopped {flavor.value} server at `{configuration.get_project_identifier()}`\n"
        )
        return commands.ExitCode.SUCCESS
    except connections.ConnectionFailure:
        LOG.info("No running Pyre server to stop.\n")
        remove_socket_if_exists(socket_path)
        return commands.ExitCode.SERVER_NOT_FOUND



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for the `pyre check` command, which runs a
single-shot type check (as opposed to `pyre incremental`, which starts a
server to enable quick incremental updates.)
"""

import contextlib
import dataclasses
import json
import logging
import subprocess
import time
from pathlib import Path
from typing import Any, Dict, Iterator, List, Sequence

from .. import (
    backend_arguments,
    command_arguments,
    configuration as configuration_module,
    error,
    frontend_configuration,
)
from . import commands, incremental, start

LOG: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class CheckResult:
    """
    Data structure for storing the result of the check command. We have this so we can call
    check from other scripts and retrieve the raw error objects, rather than having to dump them
    to stdout only to read them back in.
    """

    exit_code: commands.ExitCode
    errors: List[error.Error]


@dataclasses.dataclass(frozen=True)
class Arguments:
    """
    Data structure for configuration options the backend check command can recognize.
    Need to keep in sync with `source/command/checkCommand.ml`
    """

    base_arguments: backend_arguments.BaseArguments

    additional_logging_sections: Sequence[str] = dataclasses.field(default_factory=list)
    show_error_traces: bool = False
    strict: bool = False

    def serialize(self) -> Dict[str, Any]:
        return {
            **self.base_arguments.serialize(),
            "additional_logging_sections": self.additional_logging_sections,
            "show_error_traces": self.show_error_traces,
            "strict": self.strict,
        }


def create_check_arguments(
    configuration: frontend_configuration.Base,
    check_arguments: command_arguments.CheckArguments,
) -> Arguments:
    """
    Translate client configurations to backend check configurations.

    This API is not pure since it needs to access filesystem to filter out
    nonexistent directories. It is idempotent though, since it does not alter
    any filesystem state.
    """
    source_paths = backend_arguments.get_source_path_for_check(
        configuration,
        kill_buck_after_build=False,
        number_of_buck_threads=None,
    )

    logging_sections = check_arguments.logging_sections
    additional_logging_sections = (
        [] if logging_sections is None else logging_sections.split(",")
    )
    if check_arguments.noninteractive:
        additional_logging_sections.append("-progress")

    log_directory = configuration.get_log_directory()
    profiling_output = (
        backend_arguments.get_profiling_log_path(log_directory)
        if check_arguments.enable_profiling
        else None
    )
    memory_profiling_output = (
        backend_arguments.get_profiling_log_path(log_directory)
        if check_arguments.enable_memory_profiling
        else None
    )

    logger = configuration.get_remote_logger()
    remote_logging = (
        backend_arguments.RemoteLogging(
            logger=logger, identifier=check_arguments.log_identifier or ""
        )
        if logger is not None
        else None
    )

    return Arguments(
        base_arguments=backend_arguments.BaseArguments(
            log_path=str(log_directory),
            global_root=str(configuration.get_global_root()),
            checked_directory_allowlist=backend_arguments.get_checked_directory_allowlist(
                configuration, source_paths
            ),
            checked_directory_blocklist=(configuration.get_ignore_all_errors()),
            debug=check_arguments.debug,
            enable_readonly_analysis=configuration.get_enable_readonly_analysis(),
            enable_strict_override_check=configuration.get_enable_strict_override_check(),
            enable_strict_any_check=configuration.get_enable_strict_any_check(),
            enable_unawaited_awaitable_analysis=(
                configuration.get_enable_unawaited_awaitable_analysis()
            ),
            excludes=configuration.get_excludes(),
            extensions=configuration.get_valid_extension_suffixes(),
            include_suppressed_errors=configuration.get_include_suppressed_errors(),
            relative_local_root=configuration.get_relative_local_root(),
            memory_profiling_output=memory_profiling_output,
            number_of_workers=configuration.get_number_of_workers(),
            parallel=not check_arguments.sequential,
            profiling_output=profiling_output,
            python_version=configuration.get_python_version(),
            system_platform=configuration.get_system_platform(),
            shared_memory=configuration.get_shared_memory(),
            remote_logging=remote_logging,
            search_paths=configuration.get_existent_search_paths(),
            source_paths=source_paths,
        ),
        additional_logging_sections=additional_logging_sections,
        show_error_traces=check_arguments.show_error_traces,
        strict=configuration.is_strict(),
    )


@contextlib.contextmanager
def create_check_arguments_and_cleanup(
    configuration: frontend_configuration.Base,
    check_arguments: command_arguments.CheckArguments,
) -> Iterator[Arguments]:
    arguments = create_check_arguments(configuration, check_arguments)
    try:
        yield arguments
    finally:
        # It is safe to clean up source paths after check command since
        # any created artifact directory won't be reused by other commands.
        arguments.base_arguments.source_paths.cleanup()


class InvalidCheckResponse(Exception):
    pass


def parse_type_error_response_json(response_json: object) -> List[error.Error]:
    try:
        # The response JSON is expected to have the following form:
        # `{"errors": [error_json0, error_json1, ...]}`
        if isinstance(response_json, dict):
            errors_json = response_json.get("errors", [])
            if isinstance(errors_json, list):
                return [error.Error.from_json(error_json) for error_json in errors_json]

        raise InvalidCheckResponse(
            f"Unexpected JSON response from check command: {response_json}"
        )
    except error.ErrorParsingFailure as parsing_error:
        message = f"Unexpected error JSON from check command: {parsing_error}"
        raise InvalidCheckResponse(message) from parsing_error


def parse_type_error_response(response: str) -> List[error.Error]:
    try:
        response_json = json.loads(response)
        return parse_type_error_response_json(response_json)
    except json.JSONDecodeError as decode_error:
        message = f"Cannot parse response as JSON: {decode_error}"
        raise InvalidCheckResponse(message) from decode_error


def _run_check_command(command: Sequence[str]) -> CheckResult:
    with backend_arguments.backend_log_file(prefix="pyre_check") as log_file:
        with start.background_logging(Path(log_file.name)):
            # lint-ignore: NoUnsafeExecRule
            result = subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=log_file.file,
                universal_newlines=True,
                errors="replace",
            )
            return_code = result.returncode

            # Interpretation of the return code needs to be kept in sync with
            # `source/command/checkCommand.ml`.
            if return_code == 0:
                type_errors = parse_type_error_response(result.stdout)

                exit_code = (
                    commands.ExitCode.SUCCESS
                    if len(type_errors) == 0
                    else commands.ExitCode.FOUND_ERRORS
                )
                return CheckResult(exit_code, type_errors)
            elif return_code == 2:
                LOG.error("Pyre encountered a failure within buck.")
                return CheckResult(commands.ExitCode.BUCK_INTERNAL_ERROR, [])
            elif return_code == 3:
                LOG.error("Pyre encountered an error when building the buck targets.")
                return CheckResult(commands.ExitCode.BUCK_USER_ERROR, [])

            else:
                LOG.error(
                    f"The backend check command exited with non-zero return code: {return_code}. "
                    f"This likely indicates a problem with the Pyre binary at `{command[0]}`."
                )
                # The binary had an unexpected error. Allow a short time for
                # logs to be forwarded by the background logging thread before
                # exiting, otherwise we may drop important error messages.
                time.sleep(0.5)
                return CheckResult(commands.ExitCode.FAILURE, [])


def run_check(
    configuration: frontend_configuration.Base,
    check_arguments: command_arguments.CheckArguments,
) -> CheckResult:
    start_command = configuration.get_server_start_command(download_if_needed=True)
    if start_command is None:
        raise configuration_module.InvalidConfiguration(
            "Cannot locate a Pyre binary to run."
        )
    LOG.info(f"Pyre binary is located at `{start_command.get_pyre_binary_location()}`")

    with create_check_arguments_and_cleanup(
        configuration, check_arguments
    ) as arguments:
        with backend_arguments.temporary_argument_file(arguments) as argument_file_path:
            check_command = [
                str(start_command.get_pyre_binary_location()),
                "check",
                str(argument_file_path),
            ]
            return _run_check_command(check_command)


def run(
    configuration: frontend_configuration.Base,
    check_arguments: command_arguments.CheckArguments,
) -> commands.ExitCode:
    check_result = run_check(configuration, check_arguments)
    if check_result.exit_code in (
        commands.ExitCode.SUCCESS,
        commands.ExitCode.FOUND_ERRORS,
    ):
        incremental.display_type_errors(
            check_result.errors, output=check_arguments.output
        )
    return check_result.exit_code



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from .log import (  # noqa: F401
    cleanup,
    Color,
    configured_logger,
    enable_file_logging,
    file_tailer,
    Format,
    get_input,
    get_optional_input,
    get_yes_no_input,
    initialize,
    LOG_LEVELS,
    PERFORMANCE,
    PROMPT,
    stdout,
    StreamLogger,
    SUCCESS,
    truncate,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Fixes errors provided through stdin for all sub-projects using Pyre. This
differs from fixme_single in the sense that this runs over a whole repository
(or directory containing a top-level .pyre_configuration) and operates over
sub-projects with .pyre_configuration.local files.
"""

import argparse
import logging

from pyre_extensions import override

from ..configuration import Configuration
from ..repository import Repository
from .command import CommandArguments, ErrorSource, ErrorSuppressingCommand

LOG: logging.Logger = logging.getLogger(__name__)


class FixmeAll(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        repository: Repository,
        upgrade_version: bool,
        error_source: ErrorSource,
    ) -> None:
        super().__init__(command_arguments, repository=repository)
        self._upgrade_version: bool = upgrade_version
        self._error_source: ErrorSource = error_source

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "FixmeAll":
        command_arguments = CommandArguments.from_arguments(arguments)
        return FixmeAll(
            command_arguments,
            repository=repository,
            upgrade_version=arguments.upgrade_version,
            error_source=arguments.error_source,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(FixmeAll, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "--upgrade-version",
            action="store_true",
            help="Upgrade and clean project if a version override set.",
        )
        parser.add_argument(
            "--error-source",
            choices=list(ErrorSource),
            default=ErrorSource.GENERATE,
            type=ErrorSource,
        )

    @override
    def run(self) -> None:
        project_configuration = Configuration.find_project_configuration()
        configurations = Configuration.gather_local_configurations()
        for configuration in configurations:
            self._get_and_suppress_errors(
                configuration=configuration,
                error_source=self._error_source,
                upgrade_version=self._upgrade_version,
            )
            local_root = configuration.get_directory().resolve()
            title = "{} for {}".format(
                (
                    "Update pyre version"
                    if self._upgrade_version
                    else "Suppress pyre errors"
                ),
                str(local_root.relative_to(project_configuration.parent.resolve())),
            )
            self._repository.commit_changes(commit=(not self._no_commit), title=title)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module stores all server state related fields as a single class, so all the mutable state
can be accessed in one place. Note that this state is mutable and a singleton, so users of this module
should be aware a change to this state could affect other modules that interact with this state.
"""

from __future__ import annotations

import asyncio
import dataclasses
import enum
from asyncio import Lock
from pathlib import Path
from typing import Dict, List, Optional

from typing_extensions import Final

from .. import timer
from ..language_server import protocol as lsp
from . import pyre_server_options


class ConnectionStatus(enum.Enum):
    READY = "READY"
    DISCONNECTED = "DISCONNECTED"
    NOT_CONNECTED = "NOT_CONNECTED"
    SUSPENDED = "SUSPENDED"
    BUCK_BUILDING = "BUCK_BUILDING"
    INCREMENTAL_CHECK = "INCREMENTAL_CHECK"
    STARTING = "STARTING"


@dataclasses.dataclass(frozen=True)
class DaemonStatus:
    connection_status: ConnectionStatus
    milliseconds_since_ready: float

    def as_telemetry_dict(self) -> Dict[str, float | str]:
        return {
            "server_status_before": self.connection_status.value,
            "time_since_last_ready_ms": self.milliseconds_since_ready,
        }


@dataclasses.dataclass(frozen=True)
class OpenedDocumentState:
    code: str
    is_dirty: bool = False
    pyre_code_updated: bool = False


@dataclasses.dataclass
class DaemonStatusTracker:
    _connection_status: ConnectionStatus = ConnectionStatus.NOT_CONNECTED
    _not_ready_timer: Optional[timer.Timer] = None

    def set_status(self, new_status: ConnectionStatus) -> None:
        if new_status == ConnectionStatus.READY:
            self._not_ready_timer = None
        elif self._not_ready_timer is None:
            self._not_ready_timer = timer.Timer()
        self._connection_status = new_status

    def get_status(self) -> DaemonStatus:
        return DaemonStatus(
            connection_status=self._connection_status,
            milliseconds_since_ready=(
                0
                if self._not_ready_timer is None
                else self._not_ready_timer.stop_in_millisecond()
            ),
        )

    def is_unavailable(self) -> bool:
        return self._connection_status in {
            ConnectionStatus.NOT_CONNECTED,
            ConnectionStatus.DISCONNECTED,
            ConnectionStatus.SUSPENDED,
        }


@dataclasses.dataclass
class ServerState:
    # State that can only change on config reload
    server_options: pyre_server_options.PyreServerOptions

    # Immutable States
    client_capabilities: lsp.ClientCapabilities = lsp.ClientCapabilities()

    # Mutable States
    consecutive_start_failure: int = 0
    is_user_notified_on_buck_failure: bool = False
    opened_documents: Dict[Path, OpenedDocumentState] = dataclasses.field(
        default_factory=dict
    )
    diagnostics: Dict[Path, List[lsp.Diagnostic]] = dataclasses.field(
        default_factory=dict
    )

    # The daemon status is not reassignable, but has internal mutable state
    status_tracker: Final[DaemonStatusTracker] = dataclasses.field(
        default_factory=DaemonStatusTracker
    )

    client_register_event: Optional[asyncio.Event] = None

    pyautotargets_lock: Lock = dataclasses.field(default_factory=asyncio.Lock)

    ongoing_type_checks: int = 0



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Module responsible for reading code based on position/range.
"""

import dataclasses
from typing import ClassVar, Optional

from ..language_server import protocol as lsp


@dataclasses.dataclass(frozen=True)
class SourceCodeContext:
    MAX_LINES_BEFORE_OR_AFTER: ClassVar[int] = 2500

    @staticmethod
    def from_source_and_position(
        source: str,
        position: lsp.LspPosition,
        max_lines_before_or_after: int = MAX_LINES_BEFORE_OR_AFTER,
    ) -> Optional[str]:
        lines = source.splitlines()
        line_number = position.line

        if line_number >= len(lines):
            return None

        full_document_contents = source.splitlines()
        lower_line_number = max(position.line - max_lines_before_or_after, 0)
        higher_line_number = min(
            position.line + max_lines_before_or_after + 1,
            len(full_document_contents),
        )
        return "\n".join(full_document_contents[lower_line_number:higher_line_number])

    @staticmethod
    def character_at_position(
        source: str,
        position: lsp.LspPosition,
    ) -> Optional[str]:
        lines = source.splitlines()

        if (
            position.line >= len(lines)
            or position.line < 0
            or position.character < 0
            or position.character >= len(lines[position.line])
        ):
            return None

        return lines[position.line][position.character]

    @staticmethod
    def text_at_range(
        source: str,
        text_range: lsp.LspRange,
    ) -> Optional[str]:
        start = text_range.start
        end = text_range.end
        lines = source.splitlines()

        if (
            start.line >= len(lines)
            or start.line < 0
            or start.character < 0
            or start.character >= len(lines[start.line])
        ):
            return None
        if (
            end.line >= len(lines)
            or end.line < 0
            or end.character < 0
            or end.character > len(lines[end.line])
        ):
            return None
        if start.line > end.line or (
            start.line == end.line and start.character > end.character
        ):
            return None

        if start.line == end.line:
            return lines[start.line][start.character : end.character]

        result = ""
        result += lines[start.line][start.character :]
        for line_num in range(start.line + 1, end.line):
            result += "\n" + lines[line_num]
        result += "\n" + lines[end.line][: end.character]
        return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides some utility code for handling exit status and
uncaught exceptions in Pyre commands.
"""

import enum
import signal


class ExitCode(enum.IntEnum):
    SUCCESS = 0
    FOUND_ERRORS = 1
    FAILURE = 2
    BUCK_INTERNAL_ERROR = 3
    SERVER_NOT_FOUND = 4
    INCONSISTENT_SERVER = 5
    CONFIGURATION_ERROR = 6
    BUCK_USER_ERROR = 7
    WATCHMAN_ERROR = 8
    TAINT_CONFIGURATION_ERROR = 9
    MODEL_VERIFICATION_ERROR = 10
    UNSUPPORTED_PLATFORM = 11
    CLICK_EXCEPTION = 12
    # Exit should only happen after a shutdown message
    LANGUAGE_SERVER_EXIT = 13
    # If the process exited due to a signal, this will be the negative signal number.
    SIGSEGV = -signal.SIGSEGV


class ClientException(Exception):
    exit_code: ExitCode

    def __init__(self, message: str, exit_code: ExitCode = ExitCode.FAILURE) -> None:
        super().__init__(message)
        self.exit_code = exit_code



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Fixes errors provided through stdin for a whole project using Pyre. This
differs from fixme_single and fixme_all in the sense that this does not take
.pyre_configuration.local files into account.

If your project contains sub-projects (.pyre_configuration.local files), then
consider using fixme_single or fixme_all instead.
"""

import argparse
import logging
from typing import Optional

from pyre_extensions import override

from ..configuration import Configuration
from ..errors import Errors
from ..filesystem import add_local_mode, LocalMode
from ..repository import Repository
from .command import CommandArguments, ErrorSource, ErrorSuppressingCommand

LOG: logging.Logger = logging.getLogger(__name__)


class Fixme(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        repository: Repository,
        error_source: str,
        only_fix_error_code: Optional[int] = None,
        fixme_threshold: Optional[int] = None,
    ) -> None:
        super().__init__(command_arguments, repository)
        self._error_source: str = error_source
        self._only_fix_error_code: Optional[int] = only_fix_error_code
        self._fixme_threshold: Optional[int] = fixme_threshold

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "Fixme":
        command_arguments = CommandArguments.from_arguments(arguments)
        return Fixme(
            command_arguments,
            repository=repository,
            error_source=arguments.error_source,
            fixme_threshold=arguments.fixme_threshold,
            only_fix_error_code=arguments.only_fix_error_code,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        # Make this compatible with `fixme-single`.
        super(Fixme, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "--error-source",
            type=ErrorSource,
            choices=list(ErrorSource),
            default=ErrorSource.STDIN,
        )
        parser.add_argument(
            "--only-fix-error-code",
            type=int,
            help="Only add fixmes for errors with this specific error code.",
            default=None,
        )
        parser.add_argument(
            "--fixme-threshold",
            type=int,
            help="If a file contains more than this number of errors, add a pyre-ignore-all comment.",
            default=None,
        )

    @override
    def run(self) -> None:
        if self._error_source == ErrorSource.GENERATE:
            errors = self._generate_errors()
            self._apply_suppressions(errors)

            if self._lint:
                if self._repository.format():
                    errors = self._generate_errors()
                    self._apply_suppressions(errors)
        else:
            errors = Errors.from_stdin(self._only_fix_error_code)
            fixme_threshold = self._fixme_threshold
            if fixme_threshold is None:
                self._apply_suppressions(errors)
                return

            for path, path_errors in errors.paths_to_errors.items():
                if len(path_errors) > fixme_threshold:
                    LOG.info(
                        "%d errors found in `%s`. Adding file-level ignore.",
                        len(path_errors),
                        path,
                    )
                    add_local_mode(
                        path,
                        LocalMode.IGNORE,
                        False,  # ignore empty files (False is default)
                        True,  # override existing local mode
                    )
                else:
                    self._apply_suppressions(Errors(path_errors))

    def _generate_errors(self) -> Errors:
        configuration_path = Configuration.find_project_configuration()
        configuration = Configuration(configuration_path)
        return configuration.get_errors(self._only_fix_error_code)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import ast
import logging
from logging import Logger
from typing import Callable

from pyre_extensions import TypeVarTuple
from pyre_extensions.type_variable_operators import Concatenate

Ts = TypeVarTuple("Ts")


LOG: Logger = logging.getLogger(__name__)


class UnstableAST(Exception):
    pass


def check_stable(input: str, transformed: str) -> None:
    parsed_original = ast.parse(input)
    try:
        parsed_transformed = ast.parse(transformed)
        if ast.dump(parsed_original) != ast.dump(parsed_transformed):
            raise UnstableAST("ASTs differ")
    except SyntaxError:
        raise UnstableAST("Could not parse transformed AST")


def check_stable_transformation(
    # pyre-fixme[31]: Expression `Concatenate[(str,
    #  $local_tools?pyre?tools?upgrade?ast$Ts)], str)]` is not a valid type.
    # pyre-fixme[31]: Expression `Concatenate[(str,
    #  $local_tools?pyre?tools?upgrade?ast$Ts)], str)]` is not a valid type.
    transform: "Callable[Concatenate[str, Ts], str]",
    # pyre-fixme[31]: Expression `Concatenate[(str,
    #  $local_tools?pyre?tools?upgrade?ast$Ts)], str)]` is not a valid type.
) -> "Callable[Concatenate[str, Ts], str]":
    # pyre-fixme[11]: Annotation `Ts` is not defined as a type.
    def wrapper(input: str, *args: Ts) -> str:
        transformed = transform(input, *args)
        check_stable(input, transformed)
        return transformed

    return wrapper



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains logic for handling the json rpc protocol, which
is used for Language Server Protocol messages.
"""

import abc
import dataclasses
import json
from enum import Enum
from json.decoder import JSONDecodeError
from typing import Any, Dict, Mapping, Optional, Sequence, Union

JSON = Dict[str, Any]
JSONRPC_VERSION = "2.0"


class LanguageServerMessageType(Enum):
    """Message type for an LSP warning message."""

    WARNING = 2
    INFORMATION = 3


class JSONRPCException(Exception, metaclass=abc.ABCMeta):
    """
    Base class of all jsonrpc related errors.
    """

    @abc.abstractmethod
    def error_code(self) -> int:
        raise NotImplementedError()


class ParseError(JSONRPCException):
    """
    An error occurred on the server while parsing the JSON text.
    """

    def error_code(self) -> int:
        return -32700


class InvalidRequestError(JSONRPCException):
    """
    The JSON received is not a valid Request object.
    Internally we also raise it when the JSON sent is not a valid Response object.
    """

    def error_code(self) -> int:
        return -32600


class MissingMethodFieldInRequestError(InvalidRequestError):
    """
    Special case of InvalidRequestError, used when the "method" field
    is missing in a request.

    This error type is needed because VSCode sends us many empty requests
    that have no method and we need a specific exception that makes it easy
    to ignore these messages.
    """

    pass


class MethodNotFoundError(JSONRPCException):
    """
    The method does not exist / is not available.
    """

    def error_code(self) -> int:
        return -32601


class InvalidParameterError(JSONRPCException):
    """
    Invalid method parameter(s).
    """

    def error_code(self) -> int:
        return -32602


class InternalError(JSONRPCException):
    """
    Internal JSON-RPC error.
    """

    def error_code(self) -> int:
        return -32603


class JSONRPC(abc.ABC):
    @abc.abstractmethod
    def json(self) -> JSON:
        raise NotImplementedError()

    def serialize(self) -> str:
        return json.dumps(self.json())


def _verify_json_rpc_version(json: JSON) -> None:
    json_rpc_version = json.get("jsonrpc")
    if json_rpc_version is None:
        raise InvalidRequestError(f"Required field `jsonrpc` is missing: {json}")
    if json_rpc_version != JSONRPC_VERSION:
        raise InvalidRequestError(
            f"`jsonrpc` is expected to be '2.0' but got '{json_rpc_version}'"
        )


def _parse_json_rpc_id(json: JSON) -> Union[int, str, None]:
    id = json.get("id")
    if id is not None and not isinstance(id, int) and not isinstance(id, str):
        raise InvalidRequestError(
            f"Request ID must be either an integer or string but got {id}"
        )
    return id


def _parse_json_rpc_activity_key(json: JSON) -> Optional[JSON]:
    activity_key = json.get("activityKey")
    if activity_key is None:
        return None
    elif isinstance(activity_key, dict):
        return activity_key
    else:
        raise InvalidParameterError(
            f"Cannot parse request activityKey JSON: {activity_key}"
        )


@dataclasses.dataclass(frozen=True)
class ByPositionParameters:
    values: Sequence[object] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class ByNameParameters:
    values: Mapping[str, object] = dataclasses.field(default_factory=dict)


Parameters = Union[ByPositionParameters, ByNameParameters]


@dataclasses.dataclass(frozen=True)
class Request(JSONRPC):
    method: str
    id: Union[int, str, None] = None
    activity_key: Optional[JSON] = None
    parameters: Optional[Parameters] = None

    def json(self) -> JSON:
        parameters = self.parameters
        return {
            "jsonrpc": JSONRPC_VERSION,
            "method": self.method,
            **({"id": self.id} if self.id is not None else {}),
            **(
                {"activityKey": self.activity_key}
                if self.activity_key is not None
                else {}
            ),
            **({"params": parameters.values} if parameters is not None else {}),
        }

    def extract_parameters(self) -> Parameters:
        parameters = self.parameters
        if parameters is None:
            raise InvalidRequestError(
                f"No parameters to extract for JSON-RPC {self.method} method"
            )
        return parameters

    @staticmethod
    def from_json(request_json: JSON) -> "Request":
        """
        Parse a given JSON into a JSON-RPC request.
        Raises `InvalidRequestError` and `InvalidParameterError` if the JSON
        body is malformed.
        """
        _verify_json_rpc_version(request_json)

        method = request_json.get("method")
        if method is None:
            raise MissingMethodFieldInRequestError(
                f"Required field `method` is missing: {request_json}"
            )
        if not isinstance(method, str):
            raise InvalidRequestError(
                f"`method` is expected to be a string but got {method}"
            )

        raw_parameters = request_json.get("params")
        if raw_parameters is None:
            parameters = None
        elif isinstance(raw_parameters, list):
            parameters = ByPositionParameters(raw_parameters)
        elif isinstance(raw_parameters, dict):
            parameters = ByNameParameters(raw_parameters)
        else:
            raise InvalidParameterError(
                f"Cannot parse request parameter JSON: {raw_parameters}"
            )

        id = _parse_json_rpc_id(request_json)
        activity_key = _parse_json_rpc_activity_key(request_json)
        return Request(
            method=method, id=id, activity_key=activity_key, parameters=parameters
        )

    @staticmethod
    def from_string(request_string: str) -> "Request":
        """
        Parse a given string into a JSON-RPC request.
        - Raises `ParseError` if json parsing fails.
        - Raises `InvalidRequestError` and `InvalidParameterError` if the
          JSON body is malformed (in any way other than a missing `method`)
        """
        try:
            request_json = json.loads(request_string)
            return Request.from_json(request_json)
        except JSONDecodeError as error:
            message = f"Cannot parse string into JSON: {error}"
            raise ParseError(message) from error


@dataclasses.dataclass(frozen=True)
class Response(JSONRPC):
    id: Union[int, str, None]

    @staticmethod
    def from_json(response_json: JSON) -> "Response":
        """
        Parse a given JSON into a JSON-RPC response.
        Raises `InvalidRequestError` if the JSON body is malformed.
        """
        if "result" in response_json:
            return SuccessResponse.from_json(response_json)
        elif "error" in response_json:
            return ErrorResponse.from_json(response_json)
        else:
            raise InvalidRequestError(
                "Either `result` or `error` must be presented in JSON-RPC "
                + f"responses. Got {response_json}."
            )

    @staticmethod
    def from_string(response_string: str) -> "Response":
        """
        Parse a given string into a JSON-RPC response.
        Raises `ParseError` if the parsing fails. Raises `InvalidRequestError`
        if the JSON body is malformed.
        """
        try:
            response_json = json.loads(response_string)
            return Response.from_json(response_json)
        except JSONDecodeError as error:
            message = f"Cannot parse string into JSON: {error}"
            raise ParseError(message) from error


@dataclasses.dataclass(frozen=True)
class SuccessResponse(Response):
    result: object
    activity_key: Optional[JSON] = None

    def json(self) -> JSON:
        return {
            "jsonrpc": JSONRPC_VERSION,
            **({"id": self.id} if self.id is not None else {}),
            **(
                {"activityKey": self.activity_key}
                if self.activity_key is not None
                else {}
            ),
            "result": self.result,
        }

    @staticmethod
    def from_json(response_json: JSON) -> "SuccessResponse":
        """
        Parse a given JSON into a JSON-RPC success response.
        Raises `InvalidRequestError` if the JSON body is malformed.
        """
        _verify_json_rpc_version(response_json)

        result = response_json.get("result")
        if result is None:
            raise InvalidRequestError(
                f"Required field `result` is missing: {response_json}"
            )

        # FIXME: The `id` field is required for the respnose, but we can't
        # enforce it right now since the Pyre server may emit id-less responses
        # and that has to be fixed first.
        id = _parse_json_rpc_id(response_json)
        activity_key = _parse_json_rpc_activity_key(response_json)
        return SuccessResponse(id=id, activity_key=activity_key, result=result)


@dataclasses.dataclass(frozen=True)
class ErrorResponse(Response):
    code: int
    message: str = ""
    data: Optional[object] = None
    activity_key: Optional[JSON] = None

    def json(self) -> JSON:
        return {
            "jsonrpc": JSONRPC_VERSION,
            **({"id": self.id} if self.id is not None else {}),
            **(
                {"activityKey": self.activity_key}
                if self.activity_key is not None
                else {}
            ),
            "error": {
                "code": self.code,
                "message": self.message,
                **({"data": self.data} if self.data is not None else {}),
            },
        }

    @staticmethod
    def from_json(response_json: JSON) -> "ErrorResponse":
        """
        Parse a given JSON into a JSON-RPC error response.
        Raises `InvalidRequestError` if the JSON body is malformed.
        """
        _verify_json_rpc_version(response_json)

        error = response_json.get("error")
        if error is None:
            raise InvalidRequestError(
                f"Required field `error` is missing: {response_json}"
            )
        if not isinstance(error, dict):
            raise InvalidRequestError(f"`error` must be a dict but got {error}")

        code = error.get("code")
        if code is None:
            raise InvalidRequestError(
                f"Required field `error.code` is missing: {response_json}"
            )
        if not isinstance(code, int):
            raise InvalidRequestError(
                f"`error.code` is expected to be an int but got {code}"
            )

        message = error.get("message", "")
        if not isinstance(message, str):
            raise InvalidRequestError(
                f"`error.message` is expected to be a string but got {message}"
            )

        data = error.get("data")
        # FIXME: The `id` field is required for the respnose, but we can't
        # enforce it right now since the Pyre server may emit id-less responses
        # and that has to be fixed first.
        id = _parse_json_rpc_id(response_json)
        activity_key = _parse_json_rpc_activity_key(response_json)
        return ErrorResponse(
            id=id, activity_key=activity_key, code=code, message=message, data=data
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides a series of dataclasses and enums
that describe command-line arguments for invoking Pyre.

The `CommandArguments` dataclass represents top-level
arguments, whereas the other classes describe arguments
specific to subcommands.

For example, in the invocation
```
pyre --noninteractive start --terminal
```
the `--noninteractive` flag will determine the
`CommandArguments.noninteractive` field, whereas the `--terminal` flag will
determine the `StartArguments.terminal` field.
"""

from __future__ import annotations

import enum
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from . import dataclasses_merge
from .identifiers import PyreFlavor

TEXT: str = "text"
JSON: str = "json"
SARIF: str = "sarif"


class ProfileOutput(enum.Enum):
    _value_: str
    TRACE_EVENT = "trace_event"
    COLD_START_PHASES = "cold_start_phases"
    INCREMENTAL_UPDATES = "incremental_updates"
    TAINT = "taint"
    INDIVIDUAL_TABLE_SIZES = "individual_table_sizes"
    TOTAL_SHARED_MEMORY_SIZE_OVER_TIME = "total_shared_memory_size_over_time"
    TOTAL_SHARED_MEMORY_SIZE_OVER_TIME_GRAPH = (
        "total_shared_memory_size_over_time_graph"  # noqa B950
    )

    def __str__(self) -> str:
        return self.value


class VersionKind(str, enum.Enum):
    _value_: str
    NONE = "none"
    CLIENT = "client"
    CLIENT_AND_BINARY = "client_and_binary"


class MissingFlowsKind(str, enum.Enum):
    _value_: str
    OBSCURE = "obscure"
    TYPE = "type"


class TaintOutputFormat(str, enum.Enum):
    _value_: str
    JSON = "json"
    SHARDED_JSON = "sharded-json"


@dataclass(frozen=True)
class CommandArguments:
    local_configuration: Optional[str] = None
    version: VersionKind = VersionKind.NONE
    debug: bool = False
    sequential: bool = False
    strict: bool = False
    show_error_traces: bool = False
    output: str = TEXT
    enable_profiling: bool = False
    enable_memory_profiling: bool = False
    noninteractive: bool = False
    logging_sections: Optional[str] = None
    log_identifier: Optional[str] = None
    logger: Optional[str] = None
    no_logger: bool = False
    targets: List[str] = field(default_factory=list)
    source_directories: List[str] = field(default_factory=list)
    only_check_paths: List[str] = field(default_factory=list)
    buck_mode: Optional[str] = None
    no_saved_state: bool = False
    search_path: List[str] = field(default_factory=list)
    optional_search_path: List[str] = field(default_factory=list)
    binary: Optional[str] = None
    exclude: List[str] = field(default_factory=list)
    typeshed: Optional[str] = None
    save_initial_state_to: Optional[str] = None
    load_initial_state_from: Optional[str] = None
    changed_files_path: Optional[str] = None
    dot_pyre_directory: Optional[Path] = None
    isolation_prefix: Optional[str] = None
    python_version: Optional[str] = None
    system_platform: Optional[str] = None
    shared_memory_heap_size: Optional[int] = None
    shared_memory_dependency_table_power: Optional[int] = None
    shared_memory_hash_table_power: Optional[int] = None
    number_of_workers: Optional[int] = None
    max_number_of_workers: Optional[int] = None
    use_buck2: Optional[bool] = None
    enable_unawaited_awaitable_analysis: Optional[bool] = None
    include_suppressed_errors: Optional[bool] = None


@dataclass(frozen=True)
class StartArguments:
    changed_files_path: Optional[str] = None
    debug: bool = False
    enable_memory_profiling: bool = False
    enable_profiling: bool = False
    flavor: PyreFlavor = PyreFlavor.CLASSIC
    load_initial_state_from: Optional[str] = None
    _log_identifier: Optional[str] = None
    logging_sections: Optional[str] = None
    no_saved_state: bool = False
    no_watchman: bool = False
    noninteractive: bool = False
    save_initial_state_to: Optional[str] = None
    sequential: bool = False
    show_error_traces: bool = False
    store_type_check_resolution: bool = False
    terminal: bool = False
    wait_on_initialization: bool = False
    skip_initial_type_check: bool = False
    use_lazy_module_tracking: bool = False
    analyze_external_sources: bool = False
    number_of_buck_threads: Optional[int] = None

    @staticmethod
    def create(
        command_argument: CommandArguments,
        flavor: PyreFlavor = PyreFlavor.CLASSIC,
        no_watchman: bool = False,
        store_type_check_resolution: bool = False,
        wait_on_initialization: bool = False,
        terminal: bool = False,
        skip_initial_type_check: bool = False,
        use_lazy_module_tracking: bool = False,
        analyze_external_sources: bool = False,
        number_of_buck_threads: Optional[int] = None,
    ) -> StartArguments:
        return StartArguments(
            changed_files_path=command_argument.changed_files_path,
            debug=command_argument.debug,
            enable_memory_profiling=command_argument.enable_memory_profiling,
            enable_profiling=command_argument.enable_profiling,
            flavor=flavor,
            load_initial_state_from=command_argument.load_initial_state_from,
            _log_identifier=command_argument.log_identifier,
            logging_sections=command_argument.logging_sections,
            no_saved_state=command_argument.no_saved_state,
            no_watchman=no_watchman,
            noninteractive=command_argument.noninteractive,
            save_initial_state_to=command_argument.save_initial_state_to,
            sequential=command_argument.sequential,
            show_error_traces=command_argument.show_error_traces,
            store_type_check_resolution=store_type_check_resolution,
            terminal=terminal,
            wait_on_initialization=wait_on_initialization,
            skip_initial_type_check=skip_initial_type_check,
            use_lazy_module_tracking=use_lazy_module_tracking,
            analyze_external_sources=analyze_external_sources,
            number_of_buck_threads=number_of_buck_threads,
        )

    def get_log_identifier(self) -> str:
        """
        If a log identifier was manually set (this is usually done specifically
        to isolate telemetry, e.g. when running a performance experiment), we
        use that.

        Otherwise, we use the flavor. This keeps telemetry from various kinds
        of language servers separate so that our metrics can distinguish them.
        """
        if self._log_identifier is not None:
            return self._log_identifier
        else:
            return self.flavor.value


@dataclass(frozen=True)
class IncrementalArguments:
    output: str = TEXT
    no_start: bool = False
    start_arguments: StartArguments = field(default_factory=StartArguments)


@dataclass(frozen=True)
class CheckArguments:
    debug: bool = False
    enable_memory_profiling: bool = False
    enable_profiling: bool = False
    log_identifier: Optional[str] = None
    logging_sections: Optional[str] = None
    noninteractive: bool = False
    output: str = TEXT
    sequential: bool = False
    show_error_traces: bool = False

    @staticmethod
    def create(
        command_argument: CommandArguments,
    ) -> CheckArguments:
        return CheckArguments(
            debug=command_argument.debug,
            enable_memory_profiling=command_argument.enable_memory_profiling,
            enable_profiling=command_argument.enable_profiling,
            log_identifier=command_argument.log_identifier,
            logging_sections=command_argument.logging_sections,
            noninteractive=command_argument.noninteractive,
            output=command_argument.output,
            sequential=command_argument.sequential,
            show_error_traces=command_argument.show_error_traces,
        )


@dataclass(frozen=True)
class InferArguments:
    working_directory: Path
    annotate_attributes: bool = False
    annotate_from_existing_stubs: bool = False
    debug_infer: bool = False
    quote_annotations: bool = False
    dequalify: bool = False
    enable_memory_profiling: bool = False
    enable_profiling: bool = False
    log_identifier: Optional[str] = None
    logging_sections: Optional[str] = None
    use_future_annotations: bool = False
    in_place: bool = False
    simple_annotations: bool = False
    paths_to_modify: Optional[Set[Path]] = None
    print_only: bool = False
    read_stdin: bool = False
    sequential: bool = False
    kill_buck_after_build: bool = False
    number_of_buck_threads: Optional[int] = None


@dataclass(frozen=True)
class RageArguments:
    output: Optional[Path] = None
    server_log_count: Optional[int] = None


@dataclass(frozen=True)
class StatisticsArguments:
    paths: Optional[List[Path]] = None
    log_identifier: Optional[str] = None
    log_results: bool = False
    aggregate: bool = False
    print_summary: bool = False


@dataclass(frozen=True)
class CoverageArguments:
    working_directory: str
    paths: List[str] = field(default_factory=list)
    print_summary: bool = False


@dataclass(frozen=True)
class PysaSavedStateArguments:
    watchman_root: Optional[Path] = None
    project_name: Optional[str] = None
    preset: Optional[str] = None
    cache_critical_files: List[str] = field(default_factory=list)

    def serialize(self) -> Dict[str, Any]:
        return {
            "watchman_root": (
                str(self.watchman_root) if self.watchman_root is not None else None
            ),
            "project_name": self.project_name,
            "preset": self.preset,
            "cache_critical_files": self.cache_critical_files,
        }


@dataclasses_merge.dataclass_merge
@dataclass(frozen=True)
class AnalyzeArguments:
    debug: bool = False
    dump_call_graph: Optional[str] = None
    dump_model_query_results: Optional[str] = None
    enable_memory_profiling: bool = False
    enable_profiling: bool = False
    find_missing_flows: Optional[MissingFlowsKind] = None
    infer_self_tito: bool = True
    infer_argument_tito: bool = False
    log_identifier: Optional[str] = None
    maximum_model_source_tree_width: Optional[int] = None
    maximum_model_sink_tree_width: Optional[int] = None
    maximum_model_tito_tree_width: Optional[int] = None
    maximum_tree_depth_after_widening: Optional[int] = None
    maximum_return_access_path_width: Optional[int] = None
    maximum_return_access_path_depth_after_widening: Optional[int] = None
    maximum_tito_collapse_depth: Optional[int] = None
    maximum_tito_positions: Optional[int] = None
    maximum_overrides_to_analyze: Optional[int] = None
    maximum_tito_depth: Optional[int] = None
    maximum_trace_length: Optional[int] = None
    no_verify: bool = False
    verify_dsl: bool = False
    verify_taint_config_only: bool = False
    output: str = TEXT
    repository_root: Optional[str] = None
    rule: List[int] = field(default_factory=list)
    source: List[str] = field(default_factory=list)
    sink: List[str] = field(default_factory=list)
    transform: List[str] = field(default_factory=list)
    save_results_to: Optional[str] = None
    output_format: Optional[TaintOutputFormat] = None
    sequential: bool = False
    taint_models_path: List[str] = field(default_factory=list)
    use_cache: bool = False
    build_cache_only: bool = False
    check_invariants: bool = False
    limit_entrypoints: bool = False
    compact_ocaml_heap: bool = False
    saved_state_arguments: PysaSavedStateArguments = field(
        default_factory=PysaSavedStateArguments
    )
    compute_coverage: bool = False
    scheduler_policies_path: Optional[Path] = None
    kill_buck_after_build: bool = False
    number_of_buck_threads: Optional[int] = None
    higher_order_call_graph_max_iterations: Optional[int] = None
    maximum_target_depth: Optional[int] = None  # Used for higher order call graphs
    maximum_parameterized_targets_at_call_site: Optional[int] = (
        None  # Used for higher order call graphs
    )


@dataclass(frozen=True)
class QueryArguments:
    query: str
    no_daemon: bool
    no_validation_on_class_lookup_failure: bool
    check_arguments: CheckArguments = field(default_factory=CheckArguments)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Entrypoint for the `pyre validate-models` command.

This command will sanity check that Pysa taint models make sense for
the given project.
"""

import dataclasses
import logging
import os
from pathlib import Path
from typing import Dict, List

from .. import daemon_socket, error as error_module, frontend_configuration, identifiers
from ..language_server import connections
from . import commands, daemon_query

LOG: logging.Logger = logging.getLogger(__name__)


def _relativize_error_path(
    error: error_module.ModelVerificationError,
) -> error_module.ModelVerificationError:
    if error.path is None:
        return error

    relativized_path = Path(os.path.relpath(error.path, Path.cwd()))
    return dataclasses.replace(error, path=relativized_path)


def parse_validation_errors(
    payload: Dict[str, object],
) -> List[error_module.ModelVerificationError]:
    errors_payload = [] if "errors" not in payload else payload["errors"]
    if not isinstance(errors_payload, list):
        message = f"Invalid error payload for model validation: `{errors_payload}`."
        raise daemon_query.InvalidQueryResponse(message)

    return sorted(
        (
            _relativize_error_path(error_module.ModelVerificationError.from_json(item))
            for item in errors_payload
        ),
        key=lambda error: (error.path or Path(), error.line, error.code),
    )


def parse_validation_errors_response(
    payload: object,
) -> List[error_module.ModelVerificationError]:
    if not isinstance(payload, dict) or "response" not in payload:
        message = f"Invalid payload for model validation: `{payload}`."
        raise daemon_query.InvalidQueryResponse(message)

    response_payload = payload["response"]
    if not isinstance(response_payload, dict):
        message = (
            f"Invalid response payload for model validation: `{response_payload}`."
        )
        raise daemon_query.InvalidQueryResponse(message)

    return parse_validation_errors(response_payload)


def run(configuration: frontend_configuration.Base, output: str) -> commands.ExitCode:
    socket_path = daemon_socket.get_socket_path(
        configuration.get_project_identifier(),
        flavor=identifiers.PyreFlavor.CLASSIC,
    )
    try:
        response = daemon_query.execute_query(socket_path, "validate_taint_models()")
        validation_errors = parse_validation_errors_response(response.payload)
        error_module.print_errors(
            validation_errors, output=output, error_kind="model verification"
        )
        return commands.ExitCode.SUCCESS
    except connections.ConnectionFailure:
        LOG.warning(
            "A running Pyre server is required for models to be validated. "
            "Please run `pyre` first to set up a server."
        )
        return commands.ExitCode.SERVER_NOT_FOUND



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import ast as builtin_ast
import functools
import logging
import re
import subprocess
from enum import Enum
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional

from pyre_extensions import override

from . import ast

LOG: logging.Logger = logging.getLogger(__name__)


class LocalMode(Enum):
    IGNORE = "pyre-ignore-all-errors"
    UNSAFE = "pyre-unsafe"
    STRICT = "pyre-strict"

    def get_regex(self) -> str:
        return "^[ \t]*# *" + self.value + " *$"

    def get_comment(self) -> str:
        return "# " + self.value


class Target(NamedTuple):
    name: str
    strict: bool
    pyre: bool
    check_types: bool


class TargetCollector(builtin_ast.NodeVisitor):
    def __init__(self, pyre_only: bool) -> None:
        self._pyre_only: bool = pyre_only
        self._targets: List[Target] = []
        self._contains_strict: bool = False

    @override
    def visit_Call(self, node: builtin_ast.Call) -> None:
        target_fields = node.keywords
        name = None
        check_types = False
        is_strict = False
        uses_pyre = True
        has_typing_settings = False
        for field in target_fields:
            name = self._get_name(field, name)
            check_types = self._get_check_types(field, check_types)
            is_strict = self._get_strict(field, is_strict, uses_pyre)
            uses_pyre = self._get_uses_pyre(field, uses_pyre)
            has_typing_settings = self._get_has_typing_settings(
                field, has_typing_settings
            )

        if name and has_typing_settings and (not self._pyre_only or uses_pyre):
            self._targets.append(Target(name, is_strict, uses_pyre, check_types))
        self._contains_strict = self._contains_strict or is_strict

    def _get_name(
        self, field: builtin_ast.keyword, name: Optional[str]
    ) -> Optional[str]:
        value = field.value
        if field.arg == "name":
            if isinstance(value, builtin_ast.Str):
                return value.s
        return name

    def _get_check_types(self, field: builtin_ast.keyword, check_types: bool) -> bool:
        value = field.value
        if field.arg == "check_types":
            if isinstance(value, builtin_ast.NameConstant):
                return check_types or value.value
        return check_types

    def _get_strict(
        self, field: builtin_ast.keyword, is_strict: bool, uses_pyre: bool
    ) -> bool:
        value = field.value
        if field.arg == "check_types_options":
            if isinstance(value, builtin_ast.Str):
                return is_strict or (uses_pyre and "strict" in value.s.lower())
        elif field.arg == "typing_options":
            if isinstance(value, builtin_ast.Str):
                is_strict = is_strict or "strict" in value.s.lower()
        return is_strict

    def _get_uses_pyre(self, field: builtin_ast.keyword, uses_pyre: bool) -> bool:
        value = field.value
        if field.arg == "check_types_options":
            if isinstance(value, builtin_ast.Str):
                return uses_pyre and "mypy" not in value.s.lower()
        return uses_pyre

    def _get_has_typing_settings(
        self, field: builtin_ast.keyword, has_typing_settings: bool
    ) -> bool:
        return (
            has_typing_settings
            or field.arg == "type_checker"
            or field.arg == "typing_options"
            or field.arg == "check_types_options"
            or field.arg == "check_types"
            or field.arg == "typing"
        )

    def result(self) -> List[Target]:
        return self._targets

    def contains_strict(self) -> bool:
        return self._contains_strict


def path_exists(filename: str) -> Path:
    path = Path(filename)
    if not path.exists():
        raise ValueError(f"No file at {filename}")
    return path


def find_targets(search_root: Path, pyre_only: bool = False) -> Dict[str, List[Target]]:
    LOG.info("Finding typecheck targets in %s", search_root)
    target_files = find_files(search_root, "TARGETS")
    target_names = {}
    total_targets = 0
    for target_file in target_files:
        target_finder = TargetCollector(pyre_only)
        with open(target_file, "r") as source:
            tree = builtin_ast.parse(source.read())
            target_finder.visit(tree)
            targets = target_finder.result()
            if len(targets) > 0:
                target_names[target_file] = targets
                total_targets += len(targets)

    LOG.info(
        f"Found {total_targets} typecheck targets in {len(target_names)} "
        + "TARGETS files to analyze"
    )
    return target_names


def get_lines_with_modes(
    lines: List[str], modes: List[LocalMode]
) -> Dict[int, LocalMode]:
    lines_with_modes: Dict[int, LocalMode] = {}
    for index, line in enumerate(lines):
        for mode in modes:
            if re.match(mode.get_regex(), line):
                # add one to index, since line numbers start at 1
                lines_with_modes[index + 1] = mode
                continue
    return lines_with_modes


def remove_local_mode(path: Path, modes: List[LocalMode]) -> List[LocalMode]:
    LOG.info("Processing `%s`", str(path))
    text = path.read_text()
    if "@" "generated" in text or "@" "partially-generated" in text:
        LOG.warning("Attempting to edit generated file %s, skipping.", str(path))
        return []

    lines: List[str] = text.split("\n")
    lines_with_modes = get_lines_with_modes(lines, modes)
    lines = [line for index, line in enumerate(lines) if index not in lines_with_modes]

    new_text = "\n".join(lines)
    ast.check_stable(text, new_text)
    path.write_text(new_text)
    return list(lines_with_modes.values())


def replace_local_mode(lines: List[str], mode: LocalMode, path: Path) -> bool:
    new_lines = []
    LOG
    for line in lines:
        local_mode_line = False
        for local_mode in LocalMode:
            if re.match(local_mode.get_regex(), line):
                local_mode_line = True
        if local_mode_line:
            new_lines.append(mode.get_comment())
        else:
            new_lines.append(line)
    new_text = "\n".join(new_lines)
    path.write_text(new_text)
    return True


def add_local_mode(
    filename: str,
    mode: LocalMode,
    ignore_empty_files: bool = False,
    override: bool = False,
) -> bool:
    LOG.info("Processing `%s`", filename)
    path = Path(filename)
    text = path.read_text()

    if ignore_empty_files and len(text.strip()) == 0:
        return False

    if "@" "generated" in text or "@" "partially-generated" in text:
        LOG.warning("Attempting to edit generated file %s, skipping.", filename)
        return False

    lines: "List[str]" = text.split("\n")

    # Check if a local mode is already set.
    for line in lines:
        for local_mode in LocalMode:
            if re.match(local_mode.get_regex(), line):
                if override:
                    return replace_local_mode(lines, mode, path)
                return False

    def is_header(line: str) -> bool:
        is_comment = line.lstrip().startswith("#")
        is_pyre_ignore = (
            (
                re.match("^[ \t]*# *pyre-ignore.*$", line)
                and not re.match("^[ \t]*# *pyre-ignore-all-errors *$", line)
            )
            or re.match("^[ \t]*# *pyre-fixme.*$", line)
            or re.match("^[ \t]*# *type: ignore.*$", line)
        )
        is_autodeps_header = re.match("[ \t]*# *@manual.*$", line)
        return is_comment and not is_pyre_ignore and not is_autodeps_header

    # Add local mode.
    new_lines = []
    past_header = False
    for line in lines:
        if not past_header and not is_header(line):
            past_header = True
            if len(new_lines) != 0:
                new_lines.append("")
            new_lines.append(mode.get_comment())
        new_lines.append(line)
    new_text = "\n".join(new_lines)
    ast.check_stable(text, new_text)
    path.write_text(new_text)
    return True


class Filesystem:
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        """
        Return the list of files that match any of the patterns within root.
        If exclude is provided, files that match an exclude pattern are omitted.

        Note: The `find` command does not understand globs properly.
            e.g. 'a/*.py' will match 'a/b/c.py'
        For this reason, avoid calling this method with glob patterns.
        """

        command = ["find", "."]
        command += self._match_any(patterns)
        if exclude:
            command += ["-and", "!"]
            command += self._match_any(exclude)
        return (
            subprocess.run(command, stdout=subprocess.PIPE, cwd=root)
            .stdout.decode("utf-8")
            .split()
        )

    def _match_any(self, patterns: List[str]) -> List[str]:
        expression = []
        for pattern in patterns:
            if expression:
                expression.append("-or")
            expression.extend(["-path", f"./{pattern}"])
        return ["(", *expression, ")"]


class MercurialBackedFilesystem(Filesystem):
    @override
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        command = ["hg", "files"]
        for pattern in patterns:
            command += ["--include", pattern]
        if exclude:
            for pattern in exclude:
                command += ["--exclude", pattern]
        return (
            subprocess.run(
                command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, cwd=root
            )
            .stdout.decode("utf-8")
            .split()
        )


@functools.lru_cache(1)
def get_filesystem() -> Filesystem:
    try:
        subprocess.check_output(["hg", "status"], stderr=subprocess.DEVNULL)
        return MercurialBackedFilesystem()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return Filesystem()


def remove_non_pyre_ignores(subdirectory: Path) -> None:
    python_files = [
        subdirectory / path
        for path in get_filesystem().list(str(subdirectory), patterns=[r"**/*.py"])
    ]
    if python_files:
        LOG.info("...cleaning %s python files", len(python_files))
        remove_type_ignore_command = [
            "sed",
            "-i",
            r"s/\s*# \?type: \?ignore$//g",
        ] + [str(file) for file in python_files if file.exists()]
        subprocess.check_output(remove_type_ignore_command)


def find_files(
    directory: Path, name: str, grep_pattern: Optional[str] = None
) -> List[str]:
    grep_arguments = (
        ["-exec", "grep", "--files-with-matches", grep_pattern, "{}", "+"]
        if grep_pattern is not None
        else []
    )

    try:
        output = (
            subprocess.check_output(
                ["find", str(directory), "-name", name, *grep_arguments]
            )
            .decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError as error:
        LOG.warning(
            "Failed to find files with name `%s` in directory `%s`:\n%s",
            name,
            directory,
            error.stderr,
        )
        return []

    if output == "":
        return []
    files = output.split("\n")
    return [file.strip() for file in files]


def find_directories(directory: Path) -> List[Path]:
    return [path for path in directory.iterdir() if path.is_dir()]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import subprocess
from pathlib import Path
from typing import List, Optional


class Repository:
    MIGRATION_SUMMARY: str = (
        "Migrating buck integration to use configurations.\n "
        "For more information about this migration, please see: "
        "https://fb.workplace.com/groups/295311271085134/permalink/552700215346237/"
    )

    def commit_message(
        self,
        title: str,
        summary_override: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
    ) -> str:
        return ""

    def add_paths(self, paths: List[Path]) -> None:
        pass

    def remove_paths(self, paths: List[Path]) -> None:
        pass

    def commit_changes(
        self,
        commit: bool,
        title: Optional[str] = None,
        summary: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
        ignore_failures: bool = False,
        set_dependencies: bool = True,
    ) -> None:
        pass

    def revert_all(self, remove_untracked: bool) -> None:
        pass

    def format(self) -> bool:
        return False

    def force_format(self, paths: List[str]) -> None:
        subprocess.check_call(["pyfmt", *paths])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module describes the shape of an extension specifier in a Pyre
configuration.

We provide an hook allowing an extension to be included in the import to
support custom python runtimes - for example if the config includes
```
"extensions": [
    {"suffix": "custom_py", "include_suffix_in_module_qualifier": true}
]
```
then we would allow imports of the form
```
import package.module.custom_py
```
to resolve to `package/module.custom_py`.
"""

import dataclasses
import json

from . import exceptions


@dataclasses.dataclass(frozen=True)
class Element:
    suffix: str
    include_suffix_in_module_qualifier: bool = False

    def command_line_argument(self) -> str:
        options = ""
        if self.include_suffix_in_module_qualifier:
            options = "$" + "include_suffix_in_module_qualifier"
        return self.suffix + options

    @staticmethod
    def from_json(json: object) -> "Element":
        if isinstance(json, str):
            return Element(suffix=json, include_suffix_in_module_qualifier=False)
        elif isinstance(json, dict):
            suffix = json.get("suffix", None)
            include_suffix_in_module_qualifier = json.get(
                "include_suffix_in_module_qualifier", None
            )

            if isinstance(suffix, str) and isinstance(
                include_suffix_in_module_qualifier, bool
            ):
                return Element(
                    suffix=suffix,
                    include_suffix_in_module_qualifier=include_suffix_in_module_qualifier,
                )
        raise exceptions.InvalidConfiguration(f"Invalid extension element: {json}")

    def to_json(self) -> str:
        return json.dumps(
            {
                "suffix": self.suffix,
                "include_suffix_in_module_qualifier": self.include_suffix_in_module_qualifier,
            }
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module represents a very high level API for
all the processing that will be done by the Pyre server when a
'pyre query' command is invoked on the command line.

Queries can either operate with or without a long-standing Pyre daemon: the
modules daemon_query.py and no_daemon_query.py are responsible
for handling the low-level details of how those queries are executed.
"""

import json
import logging

from .. import (
    command_arguments,
    daemon_socket,
    frontend_configuration,
    identifiers,
    log,
)
from ..language_server import connections
from . import commands, daemon_query, no_daemon_query

LOG: logging.Logger = logging.getLogger(__name__)


HELP_MESSAGE: str = """
Possible queries:
  - attributes(class_name)
    Returns a list of attributes, including functions, for a class.
  - batch(query1(arg), query2(arg))
    Runs a batch of queries and returns a map of responses. List of given queries
    may include any combination of other valid queries except for `batch` itself.
  - callees(function)
    Calls from a given function.
  - callees_with_location(function)
    Calls from a given function, including the locations at which they are called.
  - defines(module_or_class_name)
    Returns a JSON with the signature of all defines for given module or class.
  - dump_call_graph()
    Returns a comprehensive JSON of caller -> list of callees.
  - global_leaks(function1, ...): analyzes the given function(s) and emits errors when
    global variables are mutated.
  - inline_decorators(qualified_function_name, decorators_to_skip=[decorator1, ...])
    Returns the function definition after inlining decorators.
    Allows skipping certain decorators when inlining.
  - less_or_equal(T1, T2)
    Returns whether T1 is a subtype of T2.
  - model_query(path, 'model_query_name')
    Returns in JSON a list of all models generated from the query with the name
    `model_query_name` in the directory `path`.
  - path_of_module(module)
    Gives an absolute path for `module`.
  - save_server_state('path')
    Saves Pyre's serialized state into `path`.
  - superclasses(class_name1, class_name2, ...)
    Returns a mapping of class_name to the list of superclasses for `class_name`.
    If no class name is provided, return the mapping for all classes Pyre knows about.
  - type(expression)
    Evaluates the type of `expression`.
  - types(path='path') or types('path1', 'path2', ...)
    Returns a map from each given path to a list of all types for that path.
  - validate_taint_models('optional path')
    Validates models and returns errors.
    Defaults to model path in configuration if no parameter is passed in.
"""


def _print_help_message() -> None:
    log.stdout.write(HELP_MESSAGE)


def run_query(
    configuration: frontend_configuration.Base, query_text: str
) -> commands.ExitCode:
    socket_path = daemon_socket.get_socket_path(
        configuration.get_project_identifier(),
        flavor=identifiers.PyreFlavor.CLASSIC,
    )
    try:
        if query_text == "help":
            _print_help_message()
            return commands.ExitCode.SUCCESS

        response = daemon_query.execute_query(socket_path, query_text)
        log.stdout.write(json.dumps(response.payload))
        return commands.ExitCode.SUCCESS
    except connections.ConnectionFailure:
        LOG.warning(
            "A running Pyre server is required for queries to be responded. "
            "Please run `pyre` first to set up a server."
        )
        return commands.ExitCode.SERVER_NOT_FOUND


def run(
    configuration: frontend_configuration.Base,
    query_arguments: command_arguments.QueryArguments,
) -> commands.ExitCode:
    if query_arguments.no_daemon:
        response = no_daemon_query.execute_query(
            configuration,
            query_arguments,
        )
        if response is not None:
            log.stdout.write(json.dumps(response.payload))
            return commands.ExitCode.SUCCESS
        else:
            return commands.ExitCode.FAILURE
    else:
        return run_query(configuration, query_arguments.query)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides a raw, untyped representation for daemon query
responses which handles unwrapping the outer layer of responses sent
by the classic daemon. Client code must decode the `object`-typed
`response_json` field, which requires knowing the expected shape
of the response for any given query.
"""

from __future__ import annotations

import dataclasses
import json


class InvalidQueryResponse(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class Response:
    payload: object

    @staticmethod
    def from_json(
        response_json: object,
    ) -> Response:
        if (
            isinstance(response_json, list)
            and len(response_json) > 1
            and response_json[0] == "Query"
        ):
            return Response(response_json[1])
        else:
            raise InvalidQueryResponse(
                f"Unexpected JSON response from server: {response_json}"
            )

    @staticmethod
    def parse(
        response_text: str,
    ) -> Response:
        try:
            response_json = json.loads(response_text)
            return Response.from_json(response_json)
        except json.JSONDecodeError as decode_error:
            message = f"Cannot parse response as JSON: {decode_error}"
            raise InvalidQueryResponse(message) from decode_error



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module handles logic related to naming pyre projects and
servers (which in turn impacts log locations and socket paths).
"""

import enum
from pathlib import Path
from typing import List, Optional


def get_project_identifier(
    global_root: Path,
    relative_local_root: Optional[str],
) -> str:
    project_identifier = str(global_root)
    if relative_local_root is not None:
        project_identifier = project_identifier + "//" + relative_local_root
    return project_identifier


class IllegalFlavorException(Exception):
    pass


class PyreFlavor(enum.Enum):
    """
    The pyre flavor acts as a name of a particular language-server + daemon
    pair. Its value is a factor in determining socket and log paths, which
    have to be kept separate if we are running multiple language servers
    in parallel, as well as tagging telemetry data.

    On the client side, the enum value is only really important for language
    servers, which are long-lived and in most cases need custom daemons.

    All standard pyre commands use the CLASSIC daemon.

    The flavor is restricted to be one of a few known options because we rely
    on the values for metrics and also because we depend on the names being
    short enough not to exceed socket path limits.

    """

    CLASSIC = "classic"
    SHADOW = "shadow"

    def path_suffix(self) -> str:
        return "" if self == PyreFlavor.CLASSIC else f"__{self.value}"

    @staticmethod
    def persistent_choices() -> List[str]:
        """
        Valid flavors to use for the `pyre persistent` command.
        """
        return [
            PyreFlavor.CLASSIC.value,
            PyreFlavor.SHADOW.value,
        ]

    @staticmethod
    def server_flavor_choices() -> List[str]:
        return [
            PyreFlavor.CLASSIC.value,
        ]

    def server_log_subdirectory(self) -> str:
        return self.value

    def simple_name(self) -> str:
        if self == PyreFlavor.CLASSIC:
            return "Type Checking"
        else:
            raise IllegalFlavorException(f"No simple name defined for flavor {self}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for the `pyre init-pysa` command, which:
- generate a pyre configuration if necessary
- installs project dependencies if desired so that we can analyze them
- runs pyre infer if desired which can provide extra type information
"""

import logging
import os
import subprocess
import sys
from pathlib import Path

from .. import log
from ..find_directories import find_pysa_filters_directory
from . import commands
from .initialize import (
    get_configuration_and_path,
    InitializationException,
    write_configuration,
)

LOG: logging.Logger = logging.getLogger(__name__)


def _setup_environment() -> None:
    working_directory = Path(os.getcwd()).resolve()
    # Check if venv is in working_directory
    venv_path = (
        # pyre-fixme[6]: Incompatible parameter type for `Path`, expected `str`
        Path(os.environ.get("VIRTUAL_ENV")).resolve()
        if os.environ.get("VIRTUAL_ENV", False)
        else None
    )
    if venv_path is not None and str(working_directory) in str(
        venv_path.parent.absolute()
    ):
        raise InitializationException(
            "Can't use a virtual environment in the project directory."
            " Please use one outside the project directory."
        )

    # Install dependencies
    install_dependencies = log.get_yes_no_input(
        "Would you like to install the project dependencies now?"
    )
    if install_dependencies:
        requirements_file = "requirements.txt"
        if not os.path.isfile(working_directory / requirements_file):
            requirements_file = Path(
                log.get_input(
                    f"requirements.txt file not found in `{working_directory}`"
                    " Please enter its location (eg: ./requirements-dev.txt): "
                )
            )
            requirements_file = working_directory / requirements_file
        subprocess.run(
            [sys.executable, "-m", "pip", "install", "-r", requirements_file],
            check=True,
        )
    else:
        LOG.warning(
            "You have chosen not to install dependencies."
            " Please install dependencies before running Pysa to get best results."
        )
    run_infer = log.get_yes_no_input("Would you like to generate type annotations?")
    if run_infer:
        subprocess.run(["pyre", "infer", "-i"], check=True)
    import_pysa_filter_for_sapp = log.get_yes_no_input(
        "Would you like to import filters to sapp?"
    )
    if import_pysa_filter_for_sapp:
        pysa_filter_path = find_pysa_filters_directory()
        if pysa_filter_path is not None:
            LOG.info("Importing filters to sapp")
            subprocess.run(["sapp", "filter", "import", pysa_filter_path], check=True)
        else:
            LOG.warning(
                "Couldn't infer filter directory, skipping filter import to sapp."
            )


def run(skip_environment_setup: bool) -> int:
    try:
        configuration, configuration_path = get_configuration_and_path(
            taint_models_directory_required=True
        )
        write_configuration(configuration, configuration_path)
        if not skip_environment_setup:
            _setup_environment()
        LOG.log(
            log.SUCCESS,
            "Successfully initialized an environment to run Pysa!\n"
            + "  You can now run Pysa with `pyre analyze`.",
        )
        return commands.ExitCode.SUCCESS
    except Exception as error:
        LOG.error(f"Couldn't complete Pysa initialization:\n{error}")
        return commands.ExitCode.FAILURE



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains two main responsibilities:
1. Provide all the different data models representing requests/responses for the LSP client
to/from the persistent Pyre client (such as hover, definition, etc.)

2. Provide protocols for reading requests, parsing, and writing responses to/from
the LSP client. This currently involves doing some kind of transformation to/from a JSON string
to the specific representation in class form.
"""

import asyncio
import dataclasses
import enum
import logging
import urllib
from pathlib import Path
from typing import Iterable, List, Optional, Type, TypeVar, Union

import dataclasses_json
from pyre_extensions import override

from .. import dataclasses_json_extensions as json_mixins, json_rpc
from . import connections

LOG: logging.Logger = logging.getLogger(__name__)

T = TypeVar("T", bound=json_mixins.DataclassJsonMixinWithCachedSchema)


class ServerNotInitializedError(json_rpc.JSONRPCException):
    @override
    def error_code(self) -> int:
        return -32002


class RequestFailedError(json_rpc.JSONRPCException):
    @override
    def error_code(self) -> int:
        return -32803


class RequestCancelledError(json_rpc.JSONRPCException):
    @override
    def error_code(self) -> int:
        return -32800


class ReadChannelClosedError(Exception):
    pass


async def _read_headers(
    input_channel: connections.AsyncTextReader,
) -> List[str]:
    headers = []
    header = await input_channel.read_until("\r\n")
    while header != "\r\n":
        headers.append(header)
        header = await input_channel.read_until("\r\n")
    return headers


def _get_content_length(headers: Iterable[str]) -> int:
    try:
        parts: List[str] = []
        for header in headers:
            parts = [part.strip().lower() for part in header.split(":", maxsplit=1)]
            if len(parts) <= 1:
                continue

            if parts[0] == "content-length":
                return int(parts[1])

        raise json_rpc.ParseError(f"Failed to find content length header from {parts}")
    except ValueError as error:
        raise json_rpc.ParseError(
            "Cannot parse content length into integer."
        ) from error


async def _try_read_json_rpc(
    input_channel: connections.AsyncTextReader,
) -> json_rpc.Request:
    """
    Asynchronously read a JSON-RPC request from the given input channel.

    May raise `json_rpc.ParseError`, `json_rpc.InvalidRequestError`,
    `json_prc.InvalidParameterError`, and `ReadChannelClosedError`.

    This is expected to throw errors if the editor sends empty requests,
    most callsites should use `read_nonempty_json_rpc_request`
    """
    try:
        headers = await _read_headers(input_channel)
        content_length = _get_content_length(headers)
        payload = await input_channel.read_exactly(content_length)
        return json_rpc.Request.from_string(payload)
    except asyncio.IncompleteReadError as error:
        if len(error.partial) == 0:
            raise ReadChannelClosedError(
                "Trying to read from a closed input channel"
            ) from None
        else:
            raise json_rpc.ParseError(str(error)) from None


async def read_json_rpc(
    input_channel: connections.AsyncTextReader,
) -> json_rpc.Request:
    """
    Read a JSON-RPC request from the given input channel. Ignores
    any requests that are missing the "method" field, which is needed
    because VSCode often sends empty requests.

    May raise `json_rpc.ParseError`, `json_rpc.InvalidRequestError`,
    `json_prc.InvalidParameterError`, and `ReadChannelClosedError`.
    """
    while True:
        try:
            return await _try_read_json_rpc(input_channel)
        except json_rpc.MissingMethodFieldInRequestError:
            continue


def json_rpc_payload(message: json_rpc.JSONRPC) -> str:
    payload = message.serialize()
    return f"Content-Length: {len(payload)}\r\n\r\n{payload}"


async def write_json_rpc(
    output_channel: connections.AsyncTextWriter,
    response: json_rpc.JSONRPC,
) -> None:
    """
    Asynchronously write a JSON-RPC response to the given output channel.
    """
    await output_channel.write(json_rpc_payload(response))


async def write_json_rpc_ignore_connection_error(
    output_channel: connections.AsyncTextWriter,
    response: json_rpc.JSONRPC,
) -> None:
    """
    Asynchronously write a JSON-RPC response to the given output channel, and ignore
    any `ConnectionError` that occurred.
    """
    try:
        await write_json_rpc(output_channel, response)
    except ConnectionError as error:
        LOG.info(f"Ignoring connection error while writing JSON RPC. Error: {error}")


def _parse_parameters(parameters: json_rpc.Parameters, target: Type[T]) -> T:
    """
    Parse the given JSON-RPC parameters into specified LSP parameters.
    Raise `json_rpc.InvalidRequestError`on parsing failure.
    """
    if not isinstance(parameters, json_rpc.ByNameParameters):
        raise json_rpc.InvalidRequestError(
            "Parameters for LSP requests must be passed by name"
        )
    try:
        # pyre-ignore[6, 7]: Imprecise typing of `load()`
        return target.cached_schema().load(parameters.values)
    except (KeyError, ValueError, dataclasses_json.mm.ValidationError) as error:
        raise json_rpc.InvalidRequestError(str(error)) from error


class DiagnosticTag(enum.IntEnum):
    UNNECESSARY = 1
    DEPRECATED = 2


class DiagnosticSeverity(enum.IntEnum):
    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    HINT = 4


class TextDocumentSyncKind(enum.IntEnum):
    NONE = 0
    FULL = 1
    INCREMENTAL = 2


class MessageType(enum.IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    LOG = 4


@dataclasses.dataclass(frozen=True)
class DocumentUri:
    scheme: str
    authority: str
    path: str
    query: str
    fragment: str

    def to_file_path(self) -> Optional[Path]:
        if self.scheme == "file":
            return Path(self.path)
        return None

    def unparse(self) -> str:
        return urllib.parse.urlunparse(
            (
                urllib.parse.quote(self.scheme),
                urllib.parse.quote(self.authority),
                urllib.parse.quote(self.path),
                "",
                urllib.parse.quote(self.query),
                urllib.parse.quote(self.fragment),
            )
        )

    @staticmethod
    def parse(uri: str) -> "DocumentUri":
        parsed_uri = urllib.parse.urlparse(uri)
        return DocumentUri(
            scheme=urllib.parse.unquote(parsed_uri.scheme),
            authority=urllib.parse.unquote(parsed_uri.netloc),
            path=urllib.parse.unquote(parsed_uri.path),
            query=urllib.parse.unquote(parsed_uri.query),
            fragment=urllib.parse.unquote(parsed_uri.fragment),
        )

    @staticmethod
    def from_file_path(file_path: Path) -> "DocumentUri":
        return DocumentUri(
            scheme="file", authority="", path=str(file_path), query="", fragment=""
        )


@dataclasses.dataclass(frozen=True, order=True)
class PyrePosition(json_mixins.CamlCaseAndExcludeJsonMixin):
    line: int
    character: int

    def to_lsp_position(self) -> "LspPosition":
        return LspPosition(self.line - 1, self.character)


@dataclasses.dataclass(frozen=True)
class LspPosition(json_mixins.CamlCaseAndExcludeJsonMixin):
    """LSP uses 0-indexing for lines whereas Pyre uses 1-indexing."""

    line: int
    character: int

    def to_pyre_position(self) -> "PyrePosition":
        return PyrePosition(self.line + 1, self.character)


@dataclasses.dataclass(frozen=True)
class PyreRange(json_mixins.CamlCaseAndExcludeJsonMixin):
    start: PyrePosition
    end: PyrePosition

    def to_lsp_range(self) -> "LspRange":
        return LspRange(
            start=self.start.to_lsp_position(),
            end=self.end.to_lsp_position(),
        )


@dataclasses.dataclass(frozen=True)
class LspRange(json_mixins.CamlCaseAndExcludeJsonMixin):
    start: LspPosition
    end: LspPosition


@dataclasses.dataclass(frozen=True)
class CodeDescription(json_mixins.CamlCaseAndExcludeJsonMixin):
    href: str


@dataclasses.dataclass(frozen=True)
class Diagnostic(json_mixins.CamlCaseAndExcludeJsonMixin):
    range: LspRange
    message: str
    severity: Optional[DiagnosticSeverity] = None
    code: Optional[Union[int, str]] = None
    code_description: Optional[CodeDescription] = None
    source: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Info(json_mixins.CamlCaseAndExcludeJsonMixin):
    name: str
    version: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class TextDocumentSyncClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    did_save: bool = False


@dataclasses.dataclass(frozen=True)
class PublishDiagnosticsClientTagSupport(json_mixins.CamlCaseAndExcludeJsonMixin):
    value_set: List[DiagnosticTag] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class PublishDiagnosticsClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    related_information: bool = False
    tag_support: Optional[PublishDiagnosticsClientTagSupport] = None
    version_support: bool = False


@dataclasses.dataclass(frozen=True)
class TextDocumentClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    synchronization: Optional[TextDocumentSyncClientCapabilities] = None
    publish_diagnostics: Optional[PublishDiagnosticsClientCapabilities] = None


@dataclasses.dataclass(frozen=True)
class ShowStatusRequestClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    pass


@dataclasses.dataclass(frozen=True)
class WindowClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    work_done_progress: Optional[bool] = None
    # Custom VSCode extension for status bar
    status: Optional[ShowStatusRequestClientCapabilities] = None


@dataclasses.dataclass(frozen=True)
class ClientCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: Optional[TextDocumentClientCapabilities] = None
    window: Optional[WindowClientCapabilities] = None


@dataclasses.dataclass(frozen=True)
class SaveOptions(json_mixins.CamlCaseAndExcludeJsonMixin):
    include_text: Optional[bool] = None


@dataclasses.dataclass(frozen=True)
class TextDocumentSyncOptions(json_mixins.CamlCaseAndExcludeJsonMixin):
    open_close: bool = False
    change: TextDocumentSyncKind = TextDocumentSyncKind.NONE
    save: Optional[SaveOptions] = None


@dataclasses.dataclass(frozen=True)
class ServerCapabilities(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document_sync: Optional[TextDocumentSyncOptions] = None


@dataclasses.dataclass(frozen=True)
class InitializationOptions(json_mixins.CamlCaseAndExcludeJsonMixin):
    notebook_number: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class InitializeParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    capabilities: ClientCapabilities
    process_id: Optional[int] = None
    client_info: Optional[Info] = None
    initialization_options: Optional[InitializationOptions] = None

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "InitializeParameters":
        return _parse_parameters(parameters, target=InitializeParameters)


@dataclasses.dataclass(frozen=True)
class InitializeResult(json_mixins.CamlCaseAndExcludeJsonMixin):
    capabilities: ServerCapabilities
    server_info: Optional[Info] = None


@dataclasses.dataclass(frozen=True)
class TextDocumentIdentifier(json_mixins.CamlCaseAndExcludeJsonMixin):
    uri: str

    def document_uri(self) -> DocumentUri:
        return DocumentUri.parse(self.uri)


@dataclasses.dataclass(frozen=True)
class TextDocumentItem(json_mixins.CamlCaseAndExcludeJsonMixin):
    uri: str
    language_id: str
    version: int
    text: str

    def document_uri(self) -> DocumentUri:
        return DocumentUri.parse(self.uri)


@dataclasses.dataclass(frozen=True)
class ContentChange(json_mixins.CamlCaseAndExcludeJsonMixin):
    text: str


@dataclasses.dataclass(frozen=True)
class DidOpenTextDocumentParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: TextDocumentItem

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "DidOpenTextDocumentParameters":
        return _parse_parameters(parameters, target=DidOpenTextDocumentParameters)


@dataclasses.dataclass(frozen=True)
class DidCloseTextDocumentParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: TextDocumentIdentifier

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "DidCloseTextDocumentParameters":
        return _parse_parameters(parameters, target=DidCloseTextDocumentParameters)


@dataclasses.dataclass(frozen=True)
class DidChangeTextDocumentParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: TextDocumentIdentifier
    content_changes: List[ContentChange]

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "DidChangeTextDocumentParameters":
        return _parse_parameters(parameters, target=DidChangeTextDocumentParameters)


@dataclasses.dataclass(frozen=True)
class DidSaveTextDocumentParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: TextDocumentIdentifier
    text: Optional[str] = None

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "DidSaveTextDocumentParameters":
        return _parse_parameters(parameters, target=DidSaveTextDocumentParameters)


@dataclasses.dataclass(frozen=True)
class WorkspaceConfiguration(json_mixins.CamlCaseAndExcludeJsonMixin):
    kernel_runtime_dir: List[str]


@dataclasses.dataclass(frozen=True)
class WorkspaceDidChangeConfigurationParameters(
    json_mixins.CamlCaseAndExcludeJsonMixin
):
    settings: WorkspaceConfiguration

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "WorkspaceDidChangeConfigurationParameters":
        return _parse_parameters(
            parameters, target=WorkspaceDidChangeConfigurationParameters
        )


@dataclasses.dataclass(frozen=True)
class TypeCoverageParameters(json_mixins.CamlCaseAndExcludeJsonMixin):
    text_document: TextDocumentIdentifier

    @staticmethod
    def from_json_rpc_parameters(
        parameters: json_rpc.Parameters,
    ) -> "TypeCoverageParameters":
        return _parse_parameters(parameters, target=TypeCoverageParameters)


@dataclasses.dataclass(frozen=True)
class TypeCoverageResponse(json_mixins.CamlCaseAndExcludeJsonMixin):
    """Result for nuclide-vscode-lsp coverage feature."""

    covered_percent: float
    uncovered_ranges: List[Diagnostic]
    default_message: str


@dataclasses.dataclass(frozen=True)
class LspLocation(json_mixins.CamlCaseAndExcludeJsonMixin):
    """Contains the start and end column and row for a symbol."""

    uri: str
    range: LspRange



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module describes all of the flags that control behavior of a Pyre
language server - for example, the location of the backend binary and
which features to enable.
"""

from __future__ import annotations

import dataclasses
import logging
import traceback
from pathlib import Path
from typing import Callable, Optional, Sequence

from .. import (
    backend_arguments,
    command_arguments,
    configuration as configuration_module,
    daemon_socket,
    frontend_configuration,
    identifiers,
    log_lsp_event,
)
from ..language_server import features
from . import commands, start

PyreServerOptionsReader = Callable[[], "PyreServerOptions"]
FrontendConfigurationReader = Callable[[], frontend_configuration.Base]
LOG: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class PyreServerOptions:
    server_start_command: frontend_configuration.ServerStartCommand
    project_identifier: str
    start_arguments: start.Arguments
    language_server_features: features.LanguageServerFeatures
    strict_default: bool
    excludes: Sequence[str]
    flavor: identifiers.PyreFlavor

    def get_socket_path(self) -> Path:
        return daemon_socket.get_socket_path(
            self.project_identifier,
            flavor=self.flavor,
        )

    @staticmethod
    def create_from_start_arguments(
        start_arguments: start.Arguments,
        configuration: frontend_configuration.Base,
        language_server_features: features.LanguageServerFeatures,
        flavor: identifiers.PyreFlavor,
        unsaved_changes_only: bool = False,
    ) -> PyreServerOptions:
        server_start_command = configuration.get_server_start_command(
            download_if_needed=True
        )
        if server_start_command is None:
            raise configuration_module.InvalidConfiguration(
                "Cannot locate a Pyre binary to run."
            )
        if start_arguments.watchman_root is None and not unsaved_changes_only:
            raise commands.ClientException(
                "Cannot locate a `watchman` root. Pyre's server will not function "
                "properly."
            )

        return PyreServerOptions(
            server_start_command=server_start_command,
            project_identifier=configuration.get_project_identifier(),
            start_arguments=start_arguments,
            language_server_features=language_server_features,
            strict_default=configuration.is_strict(),
            excludes=configuration.get_excludes(),
            flavor=flavor,
        )

    @staticmethod
    def create(
        start_command_argument: command_arguments.StartArguments,
        configuration: frontend_configuration.Base,
        language_server_features: features.LanguageServerFeatures,
        unsaved_changes_only: bool = False,
    ) -> PyreServerOptions:
        start_arguments = start.create_server_arguments(
            configuration,
            start_command_argument,
            False,  # kill_buck_after_build
        )

        return PyreServerOptions.create_from_start_arguments(
            start_arguments,
            configuration,
            language_server_features,
            start_command_argument.flavor,
            unsaved_changes_only,
        )

    @staticmethod
    def create_reader(
        start_command_argument: command_arguments.StartArguments,
        read_frontend_configuration: FrontendConfigurationReader,
        language_server_features: features.LanguageServerFeatures,
    ) -> PyreServerOptionsReader:
        def read() -> PyreServerOptions:
            return PyreServerOptions.create(
                start_command_argument=start_command_argument,
                configuration=read_frontend_configuration(),
                language_server_features=language_server_features,
            )

        return read


def read_server_options(
    server_options_reader: PyreServerOptionsReader,
    remote_logging: Optional[backend_arguments.RemoteLogging],
) -> "PyreServerOptions":
    try:
        LOG.info("Reading Pyre server configurations...")
        return server_options_reader()
    except Exception:
        log_lsp_event.log(
            remote_logging=remote_logging,
            event=log_lsp_event.LSPEvent.NOT_CONFIGURED,
            normals={
                "exception": traceback.format_exc(),
            },
        )
        raise



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import argparse
import dataclasses
import json
import logging
import re
import subprocess
import sys
import tempfile
from functools import cached_property
from pathlib import Path
from typing import Collection, List, Optional, Set

from .. import filesystem
from ..configuration import Configuration
from ..repository import Repository
from .command import Command

LOG: logging.Logger = logging.getLogger(__name__)
DEFAULT_INCLUDES: Collection[str] = ["**.py"]


def path_is_relative_to(file: Path, path: Path) -> bool:
    if sys.version_info >= (3, 10):
        return file.is_relative_to(path)
    else:
        # use a string hack on old Python versions; mostly needed to
        # prevent failures in github CI.
        return str(file.absolute()).startswith(str(path))


@dataclasses.dataclass(frozen=True)
class ConfigurationlessOptions:
    global_configuration: Configuration
    local_configuration: Configuration
    includes: Collection[str] = dataclasses.field(
        default_factory=lambda: DEFAULT_INCLUDES
    )
    isolation_dir: Optional[str] = dataclasses.field(default=None)

    @cached_property
    def ignore_all_errors_prefixes(self) -> Collection[Path]:
        return (
            self.global_configuration.get_resolved_ignore_path_prefixes()
            | self.local_configuration.get_resolved_ignore_path_prefixes()
        )

    @cached_property
    def exclude_patterns(self) -> Collection[re.Pattern[str]]:
        return (
            self.global_configuration.get_exclude_as_patterns()
            | self.local_configuration.get_exclude_as_patterns()
        )

    @cached_property
    def default_global_mode(self) -> filesystem.LocalMode:
        global_is_strict = (
            self.global_configuration.strict
            if self.global_configuration.strict is not None
            else False  # set default configuration strictness to UNSAFE
        )
        return (
            filesystem.LocalMode.STRICT
            if global_is_strict
            else filesystem.LocalMode.UNSAFE
        )

    @cached_property
    def default_local_mode(self) -> filesystem.LocalMode:
        default_project_strictness_setting = self.local_configuration.strict

        if default_project_strictness_setting is None:
            return self.default_global_mode
        elif default_project_strictness_setting:
            return filesystem.LocalMode.STRICT
        else:
            return filesystem.LocalMode.UNSAFE

    @cached_property
    def buck_root(self) -> Path:
        return self.get_buck_root()

    @cached_property
    def included_files(self) -> Set[Path]:
        files_to_migrate = self.get_files_from_local_configuration(
            self.local_configuration, self.buck_root
        )
        already_migrated_files = self.get_already_migrated_files(self.buck_root)

        LOG.info(f"Found {len(files_to_migrate)} files to migrate")
        LOG.debug(f"Files to migrate\n:{[str(file) for file in files_to_migrate]}")
        LOG.info(f"Found {len(already_migrated_files)} already migrated files")
        LOG.debug(
            f"Files already migrated\n:{[str(file) for file in already_migrated_files]}"
        )
        files_to_migrate -= already_migrated_files
        LOG.info(
            f"Found {len(files_to_migrate)} files to migrate, not including files already migrated"
        )

        return files_to_migrate

    def __str__(self) -> str:
        local_path = str(self.local_configuration.get_path())
        global_path = str(self.global_configuration.get_path())
        return f"ConfigurationlessOptions(local={local_path}, global={global_path})"

    def get_file_mode_to_apply(
        self,
        file: Path,
    ) -> filesystem.LocalMode:
        file = file.resolve()
        default_local_mode = self.default_local_mode
        if any(
            path_is_relative_to(file, ignore_prefix)
            for ignore_prefix in self.ignore_all_errors_prefixes
        ):
            return filesystem.LocalMode.IGNORE
        else:
            return default_local_mode

    def _get_isolation_dir(self) -> List[str]:
        if self.isolation_dir is not None:
            return ["--isolation-dir", self.isolation_dir]
        else:
            return []

    def get_buck_root(self) -> Path:
        try:
            root = Path(
                subprocess.check_output(
                    ["buck2", *self._get_isolation_dir(), "root", "--kind", "project"],
                    text=True,
                    cwd=self.local_configuration.root,
                ).strip()
            )
            LOG.info(f"buck2 root is {str(root)}")
        except FileNotFoundError as e:
            raise ValueError(
                "Could not find `buck2` executable when `targets` were specified in local configuration."
            ) from e
        return root

    @staticmethod
    def format_buck_targets_for_query(targets: Collection[str]) -> List[str]:
        targets = [
            target_expression
            for target in targets
            for target_expression in ["--target", target]
        ]
        return targets

    def _get_applicable_targets_from_wildcard_targets_buck_query(
        self, targets: Collection[str]
    ) -> Collection[str]:
        targets = self.format_buck_targets_for_query(targets)
        buck_command = [
            "buck2",
            *self._get_isolation_dir(),
            "bxl",
            "prelude//python/sourcedb/filter.bxl:filter",
            "--",
            *targets,
        ]

        LOG.info(
            f"Finding targets from wildcard expression with buck2 command: `{buck_command}`"
        )

        raw_result = subprocess.check_output(
            buck_command,
            text=True,
            cwd=self.local_configuration.root,
        )
        LOG.debug(f"Found targets:\n{raw_result}")
        result = json.loads(raw_result)

        return set(result)

    def _get_files_to_process_from_applicable_targets(
        self, applicable_targets: Collection[str], buck_root: Path
    ) -> Set[Path]:
        formatted_targets = " ".join([f"{target!r}" for target in applicable_targets])
        arguments = f"inputs( set( {formatted_targets} ) )"

        with tempfile.NamedTemporaryFile(
            "w+", prefix="pyre_configurationless_"
        ) as file:
            file.write(arguments)
            file.flush()

            buck_command = [
                "buck2",
                *self._get_isolation_dir(),
                "uquery",
                f"@{file.name}",
            ]

            LOG.info(
                f"Finding files from wildcard target expression with buck2 command: `{buck_command}`"
            )

            result = subprocess.check_output(
                buck_command,
                text=True,
                cwd=self.local_configuration.root,
            ).strip()

            LOG.debug(f"Found files from applicable targets:\n`{result}`")

            return {(buck_root / file.strip()).resolve() for file in result.split("\n")}

    def _get_files_from_wildcard_targets(
        self, wildcard_targets: Collection[str], buck_project_root: Path
    ) -> Set[Path]:
        if len(wildcard_targets) == 0:
            return set()

        applicable_targets = (
            self._get_applicable_targets_from_wildcard_targets_buck_query(
                wildcard_targets
            )
        )
        wildcard_target_files = self._get_files_to_process_from_applicable_targets(
            applicable_targets, buck_project_root
        )

        LOG.debug(
            f"Files found from wildcard target filter BXL query\n{wildcard_target_files}"
        )
        return wildcard_target_files

    def _get_sourcedb_from_buck_classic_query(
        self, targets: Collection[str]
    ) -> Optional[Path]:
        targets = self.format_buck_targets_for_query(targets)
        buck_command = [
            "buck2",
            *self._get_isolation_dir(),
            "bxl",
            "prelude//python/sourcedb/classic.bxl:build",
            "--",
            *targets,
        ]

        LOG.info(f"Finding classic targets with buck2 command: `{buck_command}`")

        raw_result = subprocess.check_output(
            buck_command,
            text=True,
            cwd=self.local_configuration.root,
        )
        result = json.loads(raw_result)
        if "db" in result:
            return Path(result["db"])
        return None

    def _get_files_from_sourcedb(
        self, sourcedb_path: Path, buck_root: Path
    ) -> Set[Path]:
        LOG.debug(f"Loading files from sourcedb at {str(sourcedb_path)}")
        with sourcedb_path.open() as file:
            loaded_sourcedb = json.load(file)

        if not isinstance(loaded_sourcedb, dict) or "build_map" not in loaded_sourcedb:
            LOG.warning(f"Malformed sourcedb at {sourcedb_path}")
            return set()

        build_map = {buck_root / file for file in loaded_sourcedb["build_map"].values()}

        if "dropped_targets" in loaded_sourcedb:
            dropped_target_paths = {
                buck_root / dropped_target["dropped_source_path"]
                for dropped_target in loaded_sourcedb["dropped_targets"].values()
            }
            build_map |= dropped_target_paths

        return {
            file
            for file in build_map
            if file.exists()
            and path_is_relative_to(file, self.local_configuration.get_path().parent)
        }

    def _get_files_from_classic_targets(
        self, classic_targets: Collection[str], buck_project_root: Path
    ) -> Set[Path]:
        if len(classic_targets) == 0:
            return set()

        sourcedb_path = self._get_sourcedb_from_buck_classic_query(classic_targets)
        if sourcedb_path is None:
            LOG.warning("No sourcedb path produced")
            return set()
        LOG.debug(f"Sourcedb path found: {sourcedb_path}")

        classic_target_files = self._get_files_from_sourcedb(
            sourcedb_path, buck_project_root
        )
        LOG.debug(
            f"Files found from classic target filter BXL query\n{classic_target_files}"
        )

        return classic_target_files

    def _get_files_to_migrate_from_targets(
        self,
        configuration_targets: List[str],
        buck_project_root: Path,
    ) -> Set[Path]:
        wildcard_targets: List[str] = [
            target for target in configuration_targets if target.endswith("...")
        ]
        classic_targets: List[str] = [
            target for target in configuration_targets if not target.endswith("...")
        ]

        wildcard_target_files = self._get_files_from_wildcard_targets(
            wildcard_targets, buck_project_root
        )
        classic_target_files = self._get_files_from_classic_targets(
            classic_targets, buck_project_root
        )

        return {
            file
            for file in wildcard_target_files | classic_target_files
            if any(file.match(pattern) for pattern in self.includes)
        }

    def _get_files_to_migrate_from_source_directories(
        self,
        source_directories: List[Path],
    ) -> Set[Path]:
        LOG.debug(f"Finding files with filesystem under {source_directories}")
        file_system = filesystem.get_filesystem()

        return {
            (source_directory / file).resolve()
            for source_directory in source_directories
            for file in file_system.list(
                str(source_directory), patterns=list(self.includes)
            )
        }

    def get_files_from_local_configuration(
        self,
        local_configuration: Configuration,
        buck_root: Path,
    ) -> Set[Path]:
        if local_configuration.targets is not None:
            files = self._get_files_to_migrate_from_targets(
                local_configuration.targets,
                buck_root,
            )
        elif local_configuration.source_directories is not None:
            source_directories = local_configuration.source_directories
            local_root = Path(local_configuration.root)
            files = self._get_files_to_migrate_from_source_directories(
                [local_root / directory for directory in source_directories],
            )
        else:
            raise ValueError(
                "Could not find `targets` or `source_directories` keys in local configuration"
            )
        LOG.debug(
            f"Found {len(files)} files in local configuration {local_configuration.get_path()}"
        )
        non_excluded_files = {
            file
            for file in files
            if not any(
                exclude_pattern.search(str(file)) is not None
                for exclude_pattern in self.exclude_patterns
            )
        }
        LOG.debug(
            f"Found {len(non_excluded_files)} in local configuration {local_configuration.get_path()}"
        )
        return non_excluded_files

    def get_already_migrated_files(
        self,
        buck_root: Path,
    ) -> Set[Path]:
        already_migrated_files: Set[Path] = set()
        nested_configurations = (
            self.local_configuration.get_nested_configuration_paths()
        )
        for configuration_path in nested_configurations:
            nested_local_configuration = Configuration(Path(configuration_path))
            migrated_files = self.get_files_from_local_configuration(
                nested_local_configuration,
                buck_root,
            )
            already_migrated_files |= migrated_files
        return already_migrated_files


class Configurationless(Command):
    def __init__(
        self,
        *,
        repository: Repository,
        path: Path,
        includes: List[str],
        commit: bool,
        force_remigrate: bool,
        isolation_dir: Optional[str] = None,
    ) -> None:
        super().__init__(repository)
        self._path: Path = path
        self._includes: List[str] = includes
        self._commit: bool = commit
        self._force_remigrate: bool = force_remigrate
        self._isolation_dir = isolation_dir

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "Configurationless":
        return Configurationless(
            repository=repository,
            path=arguments.path.resolve(),
            includes=arguments.include_file_suffixes,
            commit=(not arguments.no_commit),
            force_remigrate=arguments.force_remigrate,
            isolation_dir=arguments.isolation_dir,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(Configurationless, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "path",
            help="Path to project root with local configuration.",
            type=Path,
        )
        parser.add_argument(
            "--no-commit",
            help="Do not commit changes after completing codemod.",
            action="store_true",
        )
        parser.add_argument(
            "--include-file-suffixes",
            action="extend",
            nargs="+",
            type=str,
            default=DEFAULT_INCLUDES,
            help="The suffixes to search for and include in the codemod. Default is '**.py'.",
        )
        parser.add_argument(
            "--force-remigrate",
            help="Remigrate all files, even if they already have a mode header present",
            action="store_true",
        )
        parser.add_argument("--isolation-dir", help="Buck2 isolation dir", default=None)

    def get_options(
        self,
    ) -> ConfigurationlessOptions:
        global_configuration = Configuration(
            Configuration.find_project_configuration().resolve()
        )
        configuration_path = (self._path / ".pyre_configuration.local").resolve()
        local_configuration = Configuration(configuration_path)

        if not global_configuration.get_path().exists():
            raise ValueError(
                f"Global configuration found at {str(global_configuration.get_path())}, but path doesn't exist"
            )
        if not local_configuration.get_path().exists():
            raise ValueError(
                f"Local configuration found at {str(local_configuration.get_path())}, but path doesn't exist"
            )

        options = ConfigurationlessOptions(
            global_configuration=global_configuration,
            local_configuration=local_configuration,
        )

        LOG.info(f"Configurationless options:\n{str(options)}")

        return options

    def migrate_files(
        self, options: ConfigurationlessOptions, files_to_migrate: Set[Path]
    ) -> None:
        for file in files_to_migrate:
            file_mode = options.get_file_mode_to_apply(file)
            existing_modes = set(
                filesystem.get_lines_with_modes(
                    file.read_text().split("\n"), list(filesystem.LocalMode)
                ).values()
            )
            if file_mode is not None and (
                len(existing_modes) != 1 or file_mode not in existing_modes
            ):
                if self._force_remigrate:
                    filesystem.remove_local_mode(file, list(filesystem.LocalMode))
                filesystem.add_local_mode(str(file), file_mode, ignore_empty_files=True)

    def run(self) -> None:
        options = self.get_options()

        files_to_migrate = options.included_files

        self.migrate_files(options, files_to_migrate)

        options.local_configuration.original_contents["migration_status"] = "mode"
        options.local_configuration.write()

        self._repository.commit_changes(
            commit=self._commit,
            title=f"Configurationless migration for {str(options.local_configuration.get_path())}",
            summary="""
            We are migrating local configurations to work with
            upcoming changes we'll be making to how Pyre will do type checking
            on the CLI and in the IDE. Part of this migration requires pushing
            the type checker mode into each file that gets type checked if it's
            Pyre mode differs from the strict default in FBCode
            (i.e. `# pyre-unsafe` and `# pyre-ignore-all-errors`).
            """,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module defines a background task base class, that can be used
to start an async background task (which is basically a forked async
green thread, which exposes hooks to start, stop, and query whether
it is running but is otherwise opaque to the rest of the program).
"""

import abc
import asyncio
import logging
from typing import Optional

LOG: logging.Logger = logging.getLogger(__name__)


class Task(abc.ABC):
    @abc.abstractmethod
    async def run(self) -> None:
        raise NotImplementedError()


class TaskManager:
    """
    This class manages the lifetime of a given background task.

    It maintains one piece of internal state: the existence of an ongoing
    task, represented as an attribute of type `Optional[Future]`. When the
    attribute is not `None`, it means that the task is actively running in the
    background.
    """

    _task: Task
    _ongoing: "Optional[asyncio.Future[None]]"
    _lock: asyncio.Lock

    def __init__(self, task: Task) -> None:
        """
        Initialize a background task manager. The `task` parameter is expected
        to be a coroutine which will be executed when `ensure_task_running()`
        method is invoked.

        It is expected that the provided task does not internally swallow asyncio
        `CancelledError`. Otherwise, task shutdown may not work properly.
        """
        self._task = task
        self._ongoing = None
        self._lock = asyncio.Lock()

    async def _run_task(self) -> None:
        try:
            await self._task.run()
        except asyncio.CancelledError:
            LOG.info("Terminate background task on explicit cancelling request.")
        except Exception:
            LOG.exception("Background task unexpectedly quit")
        finally:
            self._ongoing = None

    def is_task_running(self) -> bool:
        return self._ongoing is not None

    async def ensure_task_running(self) -> None:
        """
        If the background task is not currently running, schedule it to run
        in the future by adding the task to the event loop. Note that the
        scheduled task won't get a chance to execute unless control is somehow
        yield to the event loop from the current task (e.g. via an `await` on
        something).
        """
        async with self._lock:
            if self._ongoing is None:
                # pyre-fixme[8]: Attribute has type `Optional[Future[None]]`; used
                #  as `Task[None]`.
                self._ongoing = asyncio.create_task(self._run_task())

    async def ensure_task_stop(self) -> None:
        """
        If the background task is running actively, make sure it gets stopped.
        """
        async with self._lock:
            ongoing = self._ongoing
            if ongoing is not None:
                try:
                    ongoing.cancel()
                    await ongoing
                except asyncio.CancelledError:
                    # This catch is needed when `ongoing.cancel` is called before
                    # `_run_task` gets a chance to execute.
                    LOG.info(
                        "Terminate background task on explicit cancelling request."
                    )
                finally:
                    self._ongoing = None



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module defines a timer class we can use to track and report performance.
"""

import sys
import time
from typing import Callable, Optional


class Timer:
    """
    A simple utility class to facilitate clock duration computation.

    A ``Timer`` object tracks one single state: the timestamp at which the timer
    starts. It provides various ``stop`` methods which returns the duration from
    the moment of the stop method is invoked to the moment when the timer starts.

    The benefit of using this class over using the ``time`` module in the standard
    library is that this module provides a more readable interfaces with respect to
    the resolution of the duration.
    """

    get_current_time_in_nanosecond: Callable[[], int]
    start_time: int

    def __init__(
        self,
        get_current_time_in_nanosecond: Optional[Callable[[], int]] = None,
    ) -> None:
        """
        Initialize a Timer object. Starting time will be automatically reset to
        the time when this method is invoked.

        :param get_current_time_in_nanosecond:
            The timer object will invoke this function to obtain the current
            timestamp. By default, :func:`time.perf_counter_ns` is used, which
            provides a high-precision system-wide monotonic clock that is not
            affected by NTP adjustment. The default clock should be sufficient
            for wall-clock time tracking. For other purpose such as profiling,
            the recommendation is to use alternative clocks like
            :func:`time.process_time_ns` or :func:`time.thread_time_ns`.
        """
        if get_current_time_in_nanosecond is not None:
            self.get_current_time_in_nanosecond = get_current_time_in_nanosecond
        else:
            if (sys.version_info.major, sys.version_info.minor) >= (3, 7):
                self.get_current_time_in_nanosecond = time.perf_counter_ns
            else:
                self.get_current_time_in_nanosecond = lambda: int(
                    time.perf_counter() * 1e9
                )
        self.reset()

    def reset(self) -> None:
        """
        Set the starting time of the containing timer to the current time.
        """
        self.start_time = self.get_current_time_in_nanosecond()

    def stop_in_nanosecond(self) -> int:
        """
        Return how many nanoseconds have elapsed since the start time.
        """
        return self.get_current_time_in_nanosecond() - self.start_time

    def stop_in_microsecond(self) -> float:
        """
        Return how many microseconds have elapsed since the start time.
        """
        return float(self.stop_in_nanosecond()) / 1000

    def stop_in_millisecond(self) -> float:
        """
        Return how many milliseconds have elapsed since the start time.
        """
        return float(self.stop_in_nanosecond()) / 1000000

    def stop_in_second(self) -> float:
        """
        Return how many seconds have elapsed since the start time.
        """
        return float(self.stop_in_nanosecond()) / 1000000000



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
import pathlib
import re
from logging import Logger
from pathlib import Path
from typing import Sequence

from pyre_extensions import override

from ..commands.command import Command
from ..configuration import Configuration
from ..errors import Errors
from ..filesystem import path_exists
from ..repository import Repository

LOG: Logger = logging.getLogger(__name__)


class MissingOverrideReturnAnnotations(Command):
    def __init__(self, *, repository: Repository, only_fix_error_code: int) -> None:
        super().__init__(repository)
        self._only_fix_error_code: int = only_fix_error_code

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "MissingOverrideReturnAnnotations":
        return MissingOverrideReturnAnnotations(
            repository=repository, only_fix_error_code=arguments.only_fix_error_code
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(MissingOverrideReturnAnnotations, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "--only-fix-error-code",
            type=int,
            help="Only add fixmes for errors with this specific error code.",
            default=None,
        )

    @override
    def run(self) -> None:
        errors = Errors.from_stdin(self._only_fix_error_code)
        for path, errors_for_path in errors.paths_to_errors.items():
            LOG.info("Patching errors in `%s`.", path)
            errors_for_path = sorted(
                errors_for_path, key=lambda error: error["line"], reverse=True
            )

            path = pathlib.Path(path)
            lines = path.read_text().split("\n")

            for error in errors_for_path:
                if error["code"] != 15:
                    continue
                line = error["line"] - 1

                match = re.match(r".*`(.*)`\.", error["description"])
                if not match:
                    continue
                annotation = match.groups()[0]

                # Find last closing parenthesis in after line.
                LOG.info("Looking at %d: %s", line, lines[line])
                while True:
                    if "):" in lines[line]:
                        lines[line] = lines[line].replace("):", ") -> %s:" % annotation)
                        LOG.info("%d: %s", line, lines[line])
                        break
                    else:
                        line = line + 1

            LOG.warning("Writing patched %s", str(path))
            path.write_text("\n".join(lines))


class MissingGlobalAnnotations(Command):
    def __init__(self, *, repository: Repository, only_fix_error_code: int) -> None:
        super().__init__(repository)
        self._only_fix_error_code: int = only_fix_error_code

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "MissingGlobalAnnotations":
        return MissingGlobalAnnotations(
            repository=repository, only_fix_error_code=arguments.only_fix_error_code
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(MissingGlobalAnnotations, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "--only-fix-error-code",
            type=int,
            help="Only add fixmes for errors with this specific error code.",
            default=None,
        )

    @override
    def run(self) -> None:
        errors = Errors.from_stdin(self._only_fix_error_code)
        for path, errors_for_path in errors.paths_to_errors.items():
            LOG.info("Patching errors in `%s`", path)
            errors_for_path = sorted(
                errors_for_path, key=lambda error: error["line"], reverse=True
            )

            path = pathlib.Path(path)
            lines = path.read_text().split("\n")

            for error in errors_for_path:
                if error["code"] != 5:
                    continue

                line = error["line"] - 1
                match = re.match(r".*`.*`.*`(.*)`.*", error["description"])
                if not match:
                    continue
                annotation = match.groups()[0]

                LOG.info("Looking at %d: %s", line, lines[line])
                if " =" in lines[line]:
                    lines[line] = lines[line].replace(" =", ": %s =" % annotation)
                    LOG.info("%d: %s", line, lines[line])

            path.write_text("\n".join(lines))


class SetUseBuck1(Command):
    def __init__(self, *, local_roots: Sequence[Path], repository: Repository) -> None:
        super().__init__(repository)
        self._local_roots = local_roots

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "SetUseBuck1":
        return SetUseBuck1(local_roots=arguments.local_roots, repository=repository)

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(SetUseBuck1, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "local_roots",
            help="Paths to directory with local configuration",
            type=path_exists,
            nargs="*",
        )

    @override
    def run(self) -> None:
        for local_root in self._local_roots:
            configuration = Configuration(local_root / ".pyre_configuration.local")
            configuration.set_use_buck1_if_possible()
            configuration.write()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module specifies how one can tune the backend's shared memory settings
in a .pyre_configuration file. This is needed because for large projects we
might want a larger heap and fatter hash table than the defaults.
"""

import dataclasses
from typing import Dict, Optional

from .. import dataclasses_merge


@dataclasses_merge.dataclass_merge
@dataclasses.dataclass(frozen=True)
class SharedMemory:
    heap_size: Optional[int] = None
    dependency_table_power: Optional[int] = None
    hash_table_power: Optional[int] = None

    def to_json(self) -> Dict[str, int]:
        heap_size = self.heap_size
        dependency_table_power = self.dependency_table_power
        hash_table_power = self.hash_table_power
        return {
            **({"heap_size": heap_size} if heap_size is not None else {}),
            **(
                {"dependency_table_power": dependency_table_power}
                if dependency_table_power is not None
                else {}
            ),
            **(
                {"hash_table_power": hash_table_power}
                if hash_table_power is not None
                else {}
            ),
        }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module describes a configuration option to modify how Pyre watches
some files in a project.
"""

import dataclasses
from typing import Dict

from . import exceptions


@dataclasses.dataclass(frozen=True)
class UnwatchedFiles:
    root: str
    checksum_path: str

    @staticmethod
    def from_json(json: Dict[str, object]) -> "UnwatchedFiles":
        root = json.get("root", None)
        if root is None:
            raise exceptions.InvalidConfiguration(
                "Missing `root` field in UnwatchedFiles"
            )
        if not isinstance(root, str):
            raise exceptions.InvalidConfiguration(
                "`root` field in UnwatchedFiles must be a string"
            )

        checksum_path = json.get("checksum_path", None)
        if checksum_path is None:
            raise exceptions.InvalidConfiguration(
                "Missing `checksum_path` field in UnwatchedFiles"
            )
        if not isinstance(checksum_path, str):
            raise exceptions.InvalidConfiguration(
                "`checksum_path` field in UnwatchedFiles must be a string"
            )

        return UnwatchedFiles(root=root, checksum_path=checksum_path)

    def to_json(self) -> Dict[str, str]:
        return {
            "root": self.root,
            "checksum_path": self.checksum_path,
        }


@dataclasses.dataclass(frozen=True)
class UnwatchedDependency:
    change_indicator: str
    files: UnwatchedFiles

    @staticmethod
    def from_json(json: Dict[str, object]) -> "UnwatchedDependency":
        change_indicator = json.get("change_indicator", None)
        if change_indicator is None:
            raise exceptions.InvalidConfiguration(
                "Missing `change_indicator` field in UnwatchedDependency"
            )
        if not isinstance(change_indicator, str):
            raise exceptions.InvalidConfiguration(
                "`change_indicator` field in UnwatchedDependency must be a string"
            )

        files_json = json.get("files", None)
        if files_json is None:
            raise exceptions.InvalidConfiguration(
                "Missing `files` field in UnwatchedDependency"
            )
        if not isinstance(files_json, dict):
            raise exceptions.InvalidConfiguration(
                "`files` field in UnwatchedDependency must be a dict"
            )

        return UnwatchedDependency(
            change_indicator=change_indicator,
            files=UnwatchedFiles.from_json(files_json),
        )

    def to_json(self) -> Dict[str, object]:
        return {
            "change_indicator": str(self.change_indicator),
            "files": self.files.to_json(),
        }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
from pathlib import Path
from typing import Dict, List, Optional

from pyre_extensions import override
from typing_extensions import Final

from ..configuration import Configuration
from ..filesystem import find_files
from ..repository import Repository
from .command import CommandArguments, ErrorSuppressingCommand

LOG: logging.Logger = logging.getLogger(__name__)


def consolidate_nested(
    repository: Repository, topmost: Path, nested: List[Path]
) -> None:
    total_targets = []
    for nested_configuration in nested:
        configuration = Configuration(nested_configuration)
        targets = configuration.targets
        if targets:
            total_targets.extend(targets)
            repository.remove_paths([nested_configuration])
    configuration = Configuration(topmost)
    configuration.add_targets(total_targets)
    configuration.deduplicate_targets()
    configuration.write()


class ConsolidateNestedConfigurations(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        repository: Repository,
        subdirectory: Optional[str],
    ) -> None:
        super().__init__(command_arguments, repository)
        self._subdirectory: Final[Optional[str]] = subdirectory

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "ConsolidateNestedConfigurations":
        command_arguments = CommandArguments.from_arguments(arguments)
        return ConsolidateNestedConfigurations(
            command_arguments,
            repository=repository,
            subdirectory=arguments.subdirectory,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(ConsolidateNestedConfigurations, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument("--subdirectory")

    @staticmethod
    def gather_nested_configuration_mapping(
        configurations: List[str],
    ) -> Dict[str, List[str]]:
        nested_configurations = {}
        for configuration in configurations:
            if len(nested_configurations) == 0:
                nested_configurations[configuration] = []
                continue
            inserted = False
            for topmost_configuration in nested_configurations.keys():
                existing = topmost_configuration.replace(
                    ".pyre_configuration.local", ""
                )
                current = configuration.replace(".pyre_configuration.local", "")
                if current.startswith(existing):
                    nested_configurations[topmost_configuration].append(configuration)
                    inserted = True
                    break
                elif existing.startswith(current):
                    nested_configurations[configuration] = nested_configurations[
                        topmost_configuration
                    ] + [topmost_configuration]
                    del nested_configurations[topmost_configuration]
                    inserted = True
                    break
            if not inserted:
                nested_configurations[configuration] = []
        return nested_configurations

    @override
    def run(self) -> None:
        subdirectory = self._subdirectory
        subdirectory = Path(subdirectory) if subdirectory else Path.cwd()

        # Find configurations
        configurations = sorted(find_files(subdirectory, ".pyre_configuration.local"))
        if not configurations:
            LOG.warning(
                f"Skipping consolidation. No configurations found in {subdirectory}"
            )
            return
        if len(configurations) == 1:
            configuration = configurations[0]
            LOG.warning(
                f"Skipping consolidation. Only one configuration found: {configuration}"
            )
            return

        # Gather nesting structure of configurations
        nested_configurations = self.gather_nested_configuration_mapping(configurations)
        if all(len(nested) == 0 for nested in nested_configurations.values()):
            LOG.warning(
                "Skipping consolidation. None of the configurations found are nested."
            )
            return

        # Consolidate targets
        for topmost, nested in nested_configurations.items():
            if len(nested) == 0:
                continue
            consolidate_nested(
                self._repository,
                Path(topmost),
                [Path(configuration) for configuration in nested],
            )
            configuration = Configuration(Path(topmost))
            self._get_and_suppress_errors(configuration)

        self._repository.commit_changes(
            commit=(not self._no_commit),
            title=f"Consolidate configurations in {subdirectory}",
            summary="Consolidating nested configurations.",
            set_dependencies=False,
        )
